'''
Author:    Andre Dozier
Date:      May 22, 2014
Purpose:   Analyze inputs to the SDP model for the C-BT system  
'''

import os
import numpy as np 
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as matlines
import scipy.stats as stats 
import math 
import glob

# define max water delivery 
maxdeliv = 320000.0
maxinflow = 500000.0 
binsize = 20000.0

#   gen data from truncated gamma
def gendata(thefit,maxval=maxdeliv):
    v = thefit.rvs() 
    while v > maxval or v < 0: 
        v = thefit.rvs() 
    return v

def logparams(mean, std): 
    # this formulation adapted for scipy from here: 
    #     http://en.wikipedia.org/wiki/Log-normal_distribution
    # other formulations from here: 
    #     http://ocw.mit.edu/courses/civil-and-environmental-engineering/1-151-probability-and-statistics-in-engineering-spring-2005/lecture-notes/briefnts8_relnl.pdf
    #     http://stats.stackexchange.com/questions/26608/how-do-i-estimate-the-parameters-of-a-log-normal-distribution-from-the-sample-me
    scale = mean ** 2 / np.sqrt(std**2 + mean**2) # mu
    shape = np.sqrt(np.log(1 + std**2 / mean**2)) # sigma
    return scale, shape 
 
def plot_inflow_mean_changes(total, bins, means, stds, meanlabels, figfile, colors, distn='norm'):
    # variables 
    lines = []
    inflow = (bins[:-1] + bins[1:])/2.
    
    # plot the initial histogram of actual water deliveries
    fig = plt.figure()
    probs, bins = np.histogram(total, bins=bins, normed=True)
    probs = probs/probs.sum()
    ylims = [0, probs.max() + probs.std()/5.]
    plt.bar(bins[:-1], probs, width=bins[1:]-bins[:-1], color='k', alpha=0.2, linewidth=0, edgecolor='none')
    lines.append(patches.Rectangle((0,0),1,1,color='k', alpha=0.2, linewidth=0))

    # plot each inflow mean change    
    for i in range(len(means)):
        # get the probabilies of being in each bin
        if distn=='norm': 
            probs = stats.norm.cdf(bins, means[i], stds[0])
        elif distn=='t': 
            probs = stats.t.cdf(bins, total.size, means[i], stds[0])
        elif distn=='lognorm': 
            scale,shape = logparams(means[i], stds[0])
            probs = stats.lognorm.cdf(bins,shape,loc=0,scale=scale)
        inflowprob = probs[1:] - probs[:-1]
        inflowprob[-1] += 1 - inflowprob.sum() # re-normalized to sum to 1 across bins

        # limits of x- and y-axes        
        xlims = [0, np.ceil(inflow.max()/50000.0)*50000.0]
        
        # fraction offset between bars 
        frac = 0.005 * (xlims[1]-xlims[0]) * i

        # plot 
        plt.vlines(inflow + frac, 0, inflowprob, color=colors[i], lw=3)
        
        # set labels and limits 
        plt.ylim(ylims)
        plt.xlabel('Colorado Big Thompson Project Water Deliveries (acre-feet/year)') 
        plt.ylabel('Probability')
        lines.append(matlines.Line2D(range(1), range(1), color=colors[i], linestyle='-', lw=3))
    
    # fix axis size 
    ax = plt.gca()
    for label in (ax.get_xticklabels() + ax.get_yticklabels()):
        label.set_fontsize(10) # set fontsize to 10, btter to see 
        
    # legend 
    labels = ['Historical deliveries'] + \
             ['%s fitted to historical' % distn if l == '' else \
              '%s decreased by %g%%' % (distn, float(l.replace('_Mean',''))) for l in meanlabels]
    plt.legend(lines, labels, loc='best', fontsize=10)
    
    # fix layout        
    fig.tight_layout()
     
    # save the data 
    plt.savefig(figfile) 

def plot_inflow_std_changes(total, bins, means, stds, stdlabels, figfile, colors, distn='norm'):
    # variables 
    lines = []
    inflow = (bins[:-1] + bins[1:])/2.
    
    # plot the initial histogram of actual water deliveries
    fig = plt.figure()
    probs, bins = np.histogram(total, bins=bins, normed=True)
    probs = probs/probs.sum()
    ylims = [0, probs.max()+probs.std()/5.]
    plt.bar(bins[:-1], probs, width=bins[1:]-bins[:-1], color='k', alpha=0.2, linewidth=0, edgecolor='none')
    lines.append(patches.Rectangle((0,0),1,1,color='k', alpha=0.2, linewidth=0))

    # plot each inflow mean change    
    for i in range(len(means)):
        # get the probabilies of being in each bin 
        if distn=='norm': 
            probs = stats.norm.cdf(bins, means[0], stds[i])
        elif distn=='t': 
            probs = stats.t.cdf(bins, total.size, means[0], stds[i])
        elif distn=='lognorm': 
            scale,shape = logparams(means[0], stds[i])
            probs = stats.lognorm.cdf(bins,shape,loc=0,scale=scale)
        
        inflowprob = probs[1:] - probs[:-1]
        inflowprob[-1] += 1 - inflowprob.sum() # re-normalized to sum to 1 across bins

        # limits of x- and y-axes        
        xlims = [0, np.ceil(inflow.max()/50000.0)*50000.0]
        
        # fraction offset between bars 
        frac = 0.005 * (xlims[1]-xlims[0]) * i

        # plot 
        plt.vlines(inflow + frac, 0, inflowprob, color=colors[i], lw=3)
        
        # set labels and limits 
        plt.ylim(ylims)
        plt.xlabel('Colorado Big Thompson Project Water Deliveries (acre-feet/year)') 
        plt.ylabel('Probability')
        lines.append(matlines.Line2D(range(1), range(1), color=colors[i], linestyle='-', lw=3))
    
    # fix axis size 
    ax = plt.gca()
    for label in (ax.get_xticklabels() + ax.get_yticklabels()):
        label.set_fontsize(10) # set fontsize to 10, btter to see 
        
    # legend 
    labels = ['Historical deliveries'] + \
             ['%s fitted to historical' % distn if l == '' else \
              '%s decreased by %g%%' % (distn, float(l.replace('_Std',''))) \
              for l in stdlabels]
    plt.legend(lines, labels, loc='best', fontsize=10)
    
    # fix layout        
    fig.tight_layout()
     
    # save the data 
    plt.savefig(figfile) 
    
def plot_inflow_all_changes(total, bins, means, stds, meanlabels, stdlabels, figbase, distn='norm'): 
    inflow = (bins[:-1] + bins[1:])/2. 
    for i in range(len(means)):
        for j in range(len(stds)):
            if distn=='norm': 
                nfit = stats.norm(means[i], stds[i])
            elif distn=='t': 
                nfit = stats.t(total.size, means[i], stds[i])
            elif distn=='lognorm': 
                scale,shape = logparams(means[i], stds[i])
                nfit = stats.lognorm(shape,loc=0,scale=scale)
            newdata = [gendata(thefit=nfit) for _ in range(10000)]
            
            # format output data
            print '    Plotting estimated inflowprob%s%s' % (meanlabels[i], stdlabels[j])
            fig = plt.figure()
            probs, _ = np.histogram(newdata, bins=bins, normed=True)
            inflowprob = probs * (bins[1:] - bins[:-1]) # re-normalized to sum to 1 across bins
            if i == 0 and j == 0: 
                ylims = [0, inflowprob.max() + inflowprob.std()/5.]
            plt.bar(inflow, inflowprob)
            plt.ylim(ylims)
            xlims = plt.xlim()
            frac = 0.005 * (xlims[1]-xlims[0])
            
            plt.xlabel('Estimated Inflow (acre-feet/year)') 
            plt.ylabel('Probability of entering bin')
            fig.tight_layout() 
            plt.savefig(figbase + '_inflowprob%s%s.png' % (meanlabels[i], stdlabels[j])) 
            
            # save the data 
            np.savetxt(figbase + '_inflowprob%s%s.txt' % (meanlabels[i], stdlabels[j]), inflowprob)
       
def usage(inputfile, plotvar='Total_Use', verbose=False):
    '''
    Analyzes the "inflows" to the system, which are actually heavily managed 
    according to water rights, snow pack, and storage availability on the East 
    Slope 
    '''
    
    # initialize variables 
    outdir = os.path.dirname(inputfile) + os.sep + 'figs'
    if not os.path.exists(outdir): 
        os.makedirs(outdir)
    figbase = os.path.splitext(os.path.basename(inputfile))[0]
    figbase = os.path.join(outdir, figbase) 
    print '\n\nSaving figures to "%s"' % figbase  
    
    # read the file and get the data 
    f = open(inputfile, 'r') 
    headers = f.readline().strip().split('\t')
    data = np.loadtxt(f, delimiter='\t', dtype=np.float64)
    f.close() 
    
    # parse the data 
    c = headers.index('Year')
    years = data[:, c:c+1]
    c = headers.index('Ag_Use')
    ag = data[:, c:c+1]
    c = headers.index('MI_Use')
    muni = data[:, c:c+1]
    c = headers.index('Total_Use')
    total = data[:, c:c+1]
    c = headers.index('Quota_Perc') 
    quota_perc = data[:, c:c+1]
    c = headers.index('Quota_AF') 
    quota_af = data[:, c:c+1]
    alphas = np.hstack((ag, muni)) / total
    
    # create bins 
    q = np.linspace(0, math.ceil(maxdeliv/binsize)*binsize, 100) 
    bins = np.linspace(round(total.min()/binsize)*binsize - np.mod(maxdeliv,binsize), maxdeliv, 12)
    outbins = np.linspace(0.0, maxdeliv + binsize/2., 17)
    
    # fit statistical functions to data
    #   normal 
    nparams = stats.norm.fit(total)
    normfit = stats.norm(*nparams)
    totalnorm = normfit.pdf(q)
    #   gamma 
    gparams = stats.gamma.fit(total)
    gammafit = stats.gamma(*gparams)
    totalgamma = gammafit.pdf(q)
    normdata = [gendata(normfit) for _ in range(10000)]  
    
    # plot the histogram of the totals
    print '    Plotting total historical deliveries' 
    fig = plt.figure()
    ax = fig.add_subplot(111)
    plt.hist(total, color='k', bins=outbins, alpha=0.5, normed=True)
    probs, bins, _ = plt.hist(normdata, color='r', bins=outbins, alpha=0.5, normed=True, label='Generated Norm Histogram') # , bins=bins, normed=True)
    if verbose: 
        print bins 
        print probs 
    plt.plot(q, totalnorm, label='Fitted Normal')
    plt.plot(q, totalgamma, label='Fitted Gamma')
    plt.ylim([0, plt.ylim()[1]*1.15])
    plt.xlabel('Total Deliveries (acre-feet/year)',fontsize=12)
    plt.ylabel('Density',fontsize=12)
    for label in (ax.get_xticklabels() + ax.get_yticklabels()):
        label.set_fontsize(10) # set fontsize to 10, btter to see 
    plt.legend(loc='best',fontsize=12)
    fig.tight_layout()
    plt.savefig(figbase + '_total_use.png', dpi=300)
     
    # plot the histogram of the quota percentages  
    print '    Plotting historical quota (%)' 
    fig = plt.figure() 
    plt.hist(quota_perc, normed=True)
    plt.xlabel('Quota (%)')
    plt.ylabel('Density')
    fig.tight_layout()
    plt.savefig(figbase + '_quota_perc.png', dpi=300)
 
    # plot the histogram of the quotas in acre-feet  
    print '    Plotting historical quota (AF)' 
    fig = plt.figure() 
    plt.hist(quota_af, normed=True)
    plt.xlabel('Quota (acre-feet/year)')
    plt.ylabel('Density')
    fig.tight_layout()
    plt.savefig(figbase + '_quota_af.png', dpi=300)

    # format output data 
    inflow = bins[1:] # (bins[:-1] + bins[1:])/2. 
    inflowprob = probs * (bins[1:] - bins[:-1]) # re-normalized to sum to 1 across bins
    
    # save the data 
    np.savetxt(figbase + '_inflow.txt', inflow)

    # plot changes in mean and std dev all in separate graphs 
    means = normfit.mean() * np.array([1., 0.9, 0.8])
    stds = normfit.std() * np.array([1., 1.1, 1.2]) 
    colors = ['g', 'r', 'b']
    if verbose:  
        print 'means', means
        print 'std devs', stds
    meanlabels = ['', '_Mean10', '_Mean20']
    stdlabels = ['', '_Std10', '_Std20']
    plot_inflow_all_changes(total, bins, means, stds, meanlabels, stdlabels, figbase)
    
    # plot changes in means in one graph 
    print '    Plotting mean changes in inflow probabilities'
    plot_inflow_mean_changes(total, bins, means, stds, meanlabels, figbase + 'mean_changes.png', colors)
    
    # plot changes in means in one graph 
    print '    Plotting standard deviation changes in inflow probabilities'
    plot_inflow_std_changes(total, bins, means, stds, stdlabels, figbase + 'std_changes.png', colors)
    
    # return the data  
    return inflow, inflowprob
    
def seasonalinflow(inputfile, verbose=False): 
    '''
    Analyzes the natural inflows to the C-BT system  
    '''
    
    # initialize variables 
    outdir = os.path.dirname(inputfile) + os.sep + 'figs'
    if not os.path.exists(outdir): 
        os.makedirs(outdir)
    figbase = os.path.splitext(os.path.basename(inputfile))[0]
    figbase = os.path.join(outdir, figbase) 
    print '\n\nSaving figures to "%s"' % figbase  
    
    # read the file and get the data 
    f = open(inputfile, 'r') 
    headers = f.readline().strip().split('\t')
    data = np.loadtxt(f, delimiter='\t', dtype=np.float64)
    f.close() 
    
    # parse the data 
    c = headers.index('year')
    years = data[:, c:c+1]
    c = headers.index('total')
    annual = data[:, c:c+1]*1000.
    allwinter = data[:,2:8]*1000.
    allsummer = data[:,8:14]*1000.
    winter = allwinter.sum(axis=1)
    summer = allsummer.sum(axis=1)
    if verbose: 
        print 'annual', annual
        print 'winter', winter
        print 'summer', summer
        
    def inflowplots(figbase, thedata, prefix, maxinflow=maxinflow, binsize=binsize):
        # create bins 
        q = np.linspace(0, math.ceil(maxinflow/binsize)*binsize, 100) 
        bins = np.linspace(round(thedata.min()/binsize)*binsize - np.mod(maxinflow,binsize), maxinflow, 12)
        outbins = np.linspace(0.0, maxinflow, maxinflow/binsize + 1)
        
        # fit statistical functions to data
        #   normal 
        nparams = stats.norm.fit(thedata)
        normfit = stats.norm(*nparams)
        totalnorm = normfit.pdf(q)
        #   gamma 
        lparams = stats.lognorm.fit(thedata, floc=0)
        lognormfit = stats.lognorm(*lparams)
        totallognorm = lognormfit.pdf(q)
        normdata = [gendata(lognormfit, maxval=maxinflow) for _ in range(10000)]  
        if verbose: 
            print 'nparams', nparams
            print 'lparams', lparams
            
        # plot the histogram of the totals
        print '    Plotting total historical deliveries' 
        fig = plt.figure()
        ax = fig.add_subplot(111)
        plt.hist(thedata, color='k', bins=outbins, alpha=0.5, normed=True)
        probs, bins, _ = plt.hist(normdata, color='r', bins=outbins, alpha=0.5, normed=True, label='Generated Lognormal Histogram') # , bins=bins, normed=True)
        if verbose: 
            print 'bins', bins 
            print 'probs', probs 
        plt.plot(q, totalnorm, label='Fitted Normal')
        plt.plot(q, totallognorm, label='Fitted Lognormal')
        plt.ylim([0, plt.ylim()[1]*1.15])
        plt.xlabel(prefix + ' Inflow (acre-feet)',fontsize=12)
        plt.ylabel('Density',fontsize=12)
        for label in (ax.get_xticklabels() + ax.get_yticklabels()):
            label.set_fontsize(10) # set fontsize to 10, btter to see 
        plt.legend(loc='best',fontsize=12)
        fig.tight_layout()
        plt.savefig(figbase + '_inflow.png', dpi=300)
        
        # format output data 
        inflow = bins[1:] # (bins[:-1] + bins[1:])/2. 
        inflowprob = probs * (bins[1:] - bins[:-1]) # re-normalized to sum to 1 across bins
        
        # save the data 
        np.savetxt(figbase + '_inflow.txt', inflow)
    
        # plot changes in mean and std dev all in separate graphs 
        means = normfit.mean() * np.array([1., 0.9, 0.8])
        stds = normfit.std() * np.array([1., 1.1, 1.2]) 
        colors = ['g', 'r', 'b']
        if verbose:  
            print 'means', means
            print 'std devs', stds
        meanlabels = ['', '_Mean10', '_Mean20']
        stdlabels = ['', '_Std10', '_Std20']
        plot_inflow_all_changes(thedata, bins, means, stds, meanlabels, stdlabels, figbase, distn='lognorm')
        
        # plot changes in means in one graph 
        print '    Plotting mean changes in inflow probabilities'
        plot_inflow_mean_changes(thedata, bins, means, stds, meanlabels, figbase + '_mean_changes.png', colors, distn='lognorm')
        
        # plot changes in means in one graph 
        print '    Plotting standard deviation changes in inflow probabilities'
        plot_inflow_std_changes(thedata, bins, means, stds, stdlabels, figbase + '_std_changes.png', colors, distn='lognorm')
        
        return inflow, inflowprob
    
    annualinflow, annualinflowprob = inflowplots(figbase + '_annual', annual, 'Annual', maxinflow=500000, binsize=20000)
    winterinflow, winterinflowprob = inflowplots(figbase + '_winter', winter, 'Winter', maxinflow=80000, binsize=10000) 
    summerinflow, summerinflowprob = inflowplots(figbase + '_summer', summer, 'Summer', maxinflow=460000, binsize=20000)
    
    # return the data  
    return annualinflow, annualinflowprob, winterinflow, winterinflowprob, summerinflow, summerinflowprob

def summary_seas(inputfile):
    
    # initialize variables
    print '\n\nSummarizing the inflow files into single inflow file scenarios...' 
    indir = os.path.dirname(inputfile) 
    outdir = indir + os.sep + 'figs'
    if not os.path.exists(outdir): 
        os.makedirs(outdir)
    figbasename = os.path.splitext(os.path.basename(inputfile))[0]
    figbase = os.path.join(outdir, figbasename) 

    # list inflow files 
    print '    - listing files...'
    winterinflowfile = figbase + '_winter_inflow.txt'
    winterinflowprobfiles = glob.glob(figbase + '_winter_inflowprob*.txt')
    summerinflowfile = figbase + '_summer_inflow.txt'
    summerinflowprobfiles = glob.glob(figbase + '_summer_inflowprob*.txt')
    
    # read inflow files 
    winterinflow = np.loadtxt(winterinflowfile) 
    summerinflow = np.loadtxt(summerinflowfile)
    nwinter = len(winterinflow)
    nsummer = len(summerinflow) 
    nrows = nwinter + nsummer
    ncols = 3 # Time, Inflow Bins, Inflow Probability
    
    # number of files check  
    nfiles = len(winterinflowprobfiles)
    if len(summerinflowprobfiles) != nfiles: 
        raise Exception('Ensure that the number of scenarios for the summer is equivalent to those for the winter')
    
    # combine files into output array 
    print '    - building output array...'
    outarray = np.zeros((nrows, ncols))
    outarray[nwinter:,0] = 1.0
    outarray[:nwinter, 1] = winterinflow 
    outarray[nwinter:, 1] = summerinflow 
    for i in range(nfiles):
        # load winter and summer inflow probabilities  
        winflowprob = np.loadtxt(winterinflowprobfiles[i])
        sinflowprob = np.loadtxt(summerinflowprobfiles[i])
        
        # build output array 
        outarray[:nwinter, 2] = winflowprob 
        outarray[nwinter:, 2] = sinflowprob
        
        # filter out no data 
        savearray = outarray[outarray[:,2] >= 0.0001]
        
        # save output 
        outfile = winterinflowprobfiles[i].replace(figbase + '_winter_', '')
        print '    - saving %s' % outfile      
        outdir2 = os.path.join(indir, 'seas') 
        if not os.path.exists(outdir2): 
            os.makedirs(outdir2) 
        outfile = os.path.join(outdir2, outfile)
        np.savetxt(outfile, savearray, delimiter='\t', header='Time\tInflow Bins\tInflow Probability', comments='')

def summary_annual(inputfile):
    
    # initialize variables
    print '\n\nSummarizing the inflow files into single inflow file scenarios...' 
    indir = os.path.dirname(inputfile) 
    outdir = indir + os.sep + 'figs'
    if not os.path.exists(outdir): 
        os.makedirs(outdir)
    figbasename = os.path.splitext(os.path.basename(inputfile))[0]
    figbase = os.path.join(outdir, figbasename) 

    # list inflow files 
    print '    - listing files...'
    inflowfile = figbase + '_annual_inflow.txt'
    inflowprobfiles = glob.glob(figbase + '_annual_inflowprob*.txt')
    
    # read inflow files 
    inflow = np.loadtxt(inflowfile) 
    nrows = len(inflow)
    ncols = 3 # Time, Inflow Bins, Inflow Probability
    
    # number of files check  
    nfiles = len(inflowprobfiles)
    
    # combine files into output array 
    print '    - building output array...'
    outarray = np.zeros((nrows, ncols))
    outarray[:, 1] = inflow 
    for i in range(nfiles):
        # load winter and summer inflow probabilities  
        inflowprob = np.loadtxt(inflowprobfiles[i])
        
        # build output array 
        outarray[:, 2] = inflowprob 
        
        # filter out no data 
        savearray = outarray[outarray[:, 2] >= 0.0001]
        
        # save output 
        outfile = inflowprobfiles[i].replace(figbase + '_annual_', '')
        print '    - saving %s' % outfile      
        outdir2 = os.path.join(indir, 'annual') 
        if not os.path.exists(outdir2): 
            os.makedirs(outdir2) 
        outfile = os.path.join(outdir2, outfile)
        np.savetxt(outfile, savearray, delimiter='\t', header='Time\tInflow Bins\tInflow Probability', comments='')
        
if __name__ == '__main__': 
    
    # clear all other figures 
    plt.close('all')
    
    # analyze usage 
    usage(os.path.join('data','cbt_usage.txt'))
    
    # analyze inflow
    inflowfile = os.path.join('data','cbt_USBR.txt') 
    seasonalinflow(inflowfile, verbose=False)
    
    # write input files for optimization algorithm
    summary_seas(inflowfile)
    summary_annual(inflowfile)
    
    
    
    
    
    
