# imports... get dypro in the system search path 
import sys, os
import os, time, itertools, string, csv
import dypro.dypro as dp
import numpy as np
from scipy.optimize import fsolve, minimize 

def getkey(thedict, thevalue): 
    # get the key of a dictionary from a matching value...
    for k in thedict:
        if thedict[k] == thevalue: 
            return k
    return None

def integrate(intercepts, slopes):
    # check number of values 
    nvals = len(intercepts)
    if nvals != len(slopes):
        raise Exception('Both intercepts and slopes must be a list with an equivalent number of elements.')
    
    # prepare output variables
    pieces=[0.0]
    totals=[0.0]
    
    # get the "kink" locations in the piecewise function
    for i in range(nvals):
        if i == nvals - 1: 
            def diff(q): 
                return intercepts[i]   + slopes[i]   * q
        else: 
            def diff(q): 
                return (intercepts[i+1] + slopes[i+1] * q) - \
                       (intercepts[i]   + slopes[i]   * q)
        
        # solve for where the lines cross
        q = fsolve(diff, 0)
        pieces.append(q)
        totals.append(q * (intercepts[i] + slopes[i] / 2. * q) 
                      - pieces[i] * (intercepts[i] + slopes[i] / 2. * pieces[i])
                      + totals[i])
    return np.array(pieces), np.array(totals)

def getgrid(minv, maxv, stepsize):
    if (np.mod(maxv-minv, stepsize) != 0).any():
        raise Exception('stepsize must divide exactly into max-min=%s.' % str(maxv-minv))
    return dp.LinGrid(minv, maxv, np.int32((maxv - minv) / stepsize) + 1)

class ScenarioType(): 
    FreeMarket = 0
    FreeMarketBanking = 1
    PriorApprop = 2
    PriorAppropBanking = 3
    
    @staticmethod    
    def get_name(enumval):
        return getkey(ScenarioType.__dict__, enumval)

class DecisionType(): 
    Release = 0
    Save = 1
    Withdraw = 2
    Spill = 3

    @staticmethod    
    def get_name(enumval):
        return getkey(DecisionType.__dict__, enumval)
    
class PiecewiseLinearDemand:
    '''
    A piecewise linear demand function defined by a list of intercepts and 
    slopes 
    '''    
    def __init__(self, intercepts, slopes, verbose=False):
        self.intercepts = np.array(intercepts, dtype=np.float64)
        self.slopes = np.array(slopes, dtype=np.float64) 
        self.pieces, self.totals = integrate(intercepts, slopes)
        self.nkinks = len(self.intercepts) - 1
        
        # initial calculations 
        self.opt_decision = self.pieces[-1]
        if verbose: 
            print 'Maximum profit: ', self.opt_decision
        self._get_indices = np.vectorize(self._get_index)

    def _get_index(self, q):
        i = np.where(self.pieces <= q)[0]
        if len(i) == 0: 
            return 0
        else: 
            return min(i[-1], self.nkinks)
        
    def _integ(self, q, i):
        return q * (self.intercepts[i] + self.slopes[i] / 2. * q)
    
    def fix_decision(self, q): 
        # when there is an option to decrease high water deliveries, update 
        # the decision matrix
        return np.minimum(q, self.opt_decision)
    
    def total_value(self, q):
        # get the bin where q is 
        i = self._get_indices(q)
        
        # return the integral 
        return self._integ(q, i) \
                - self._integ(self.pieces[i], i) \
                + self.totals[i]
                
    def marg_value(self, q): 
        i = self._get_indices(q)
        return np.maximum(self.intercepts[i] + self.slopes[i] * q, 0.0)

class PiecewiseLinearDemandList: 
    '''
    Combines multiple linear demand functions into one interface
    '''
    def __init__(self, funcs):
        '''
        @summary:           Constructs a PiecewiseLinearDemandList object 
        @param funcs:       A list of PieceLinearDemand objects 
        '''
        self.funcs = funcs # should be list of PieceLinearDemand
        self.n = len(funcs)
        self.indices = range(self.n)
        
        # initial calculations 
        self.opt_decision = np.asarray([self.funcs[i].opt_decision if self.funcs[i] != None else 0.0 for i in self.indices])

    def total_value(self, q):
        return np.asarray([self.funcs[i].total_value(q[i]) if self.funcs[i] != None else 0.0 for i in self.indices]).sum()
                
    def marg_value(self, q): 
        return np.asarray([self.funcs[i].marg_value(q[i]) if self.funcs[i] != None else 0.0 for i in self.indices])

    def fix_decision(self, q):
        return np.asarray([self.funcs[i].fix_decision(q[i]) if self.funcs[i] != None else 0.0 for i in self.indices])
    
class PiecewiseProblem: 
    '''
    Builds a scenario for running through the C-BT optimization problem based 
    on piecewise linear demand functions, and can have multiple periods and 
    users 
    '''
    def __init__(self,
                    # water users and demand functions
                    users, abbrev, funcs, 
                    # indexed by timestep 
                    periods, upperbounds, stepsizes, maxnumerrprints=50): 
        '''
        @summary:           Constructs a PiecewiseProblem
        @param users:       Names of users 
        @param abbrev:      Abbreviated user names 
        @param funcs:       A list of PieceLinearDemand objects
        @param periods:     A list of period names within each timestep
                            (e.g., ['Winter', 'Summer'])
        @param upperbounds: A list of bounds on decision user water deliveries 
                            for each period in the same order as "periods". 
                            Each element in the list is a list of the 
                            upper bounds with the same number of elements as 
                            the number of users. Use 0 to indicate the user does 
                            not user water in a particular period.
        @param stepsizes:   A list of stepsizes for each water user and period
                            in the same order as "periods". Each element is 
                            a list of stepsizes with the same number of elements 
                            as users. 
        @param spill_upper: The upper bound on spill during optimization at each
                            period
        @param spill_step:  The stepsize at which the spill is evaluated at each 
                            period
        '''
        
        # water users and demand functions
        self.users = users 
        self.abbrev = abbrev
        self.funcs = funcs
        self.demfxn = PiecewiseLinearDemandList(funcs)
        self.nusers = len(users)
        
        # indexed by timestep 
        self.periods = periods 
        self.nperiods = len(self.periods)
        self.initialized = False 
        self.upperbounds = np.array(upperbounds, dtype=np.float64, ndmin=2)
        self.stepsizes = np.array(stepsizes, dtype=np.float64, ndmin=2)
        
        # number of prints 
        self.maxnumerrprints = maxnumerrprints
        
    def initialize(self,# describe scenario and dir locations
                     outdir, 
                     # scenario 
                     scen_name, scen_type,  
                     # inflow 
                     inflowfile,
                     # spill
                     spill_upper, 
                     # storage 
                     storage_capacity,
                     # optional variables 
                     # spill 
                     spill_step=None, 
                     # bank  
                     bank_percentage=None, withdrawal_loss=None,
                     # alloted percentage 
                     alpha=None):
        '''
        @param upperbounds: A list of bounds on decision user water deliveries 
                            for each period in the same order as "periods". 
                            Each element in the list is a list of the 
                            upper bounds with the same number of elements as 
                            the number of users. Use 0 to indicate the user does 
                            not user water in a particular period.
        @param stepsizes:   A list of stepsizes for each water user and period
                            in the same order as "periods". Each element is 
                            a list of stepsizes with the same number of elements 
                            as users. 
        @param spill_upper: The upper bound on spill during optimization at each
                            period
        @param spill_step:  The stepsize at which the spill is evaluated at each 
                            period
        @param storage_capacity: 
                            The capacity for the reservoir 
        @param bank_percentage: 
                            The percentage of a user's allotment that he or she 
                            can store in the bank
        @param withdrawal_loss:
                            The percentage of stored water that is lost from a 
                            user's banked water when the user withdraws from 
                            his/her account
        ''' 
        
        # update instance variables 
        self.outdir = outdir
        self.scen_name = scen_name 
        self.scen_type = scen_type
        self.inflowfile = inflowfile
        self.spill_upper = np.array(spill_upper, dtype=np.float64, ndmin=1)
        if len(self.spill_upper) == 1: 
            self.spill_upper = np.repeat(self.spill_upper, self.nperiods)
        self.spill_step = spill_step if spill_step != None else self.stepsizes.min(axis=1)
        self.storage_capacity = np.array(storage_capacity, dtype=np.float64, ndmin=1)
        self.bank_percentage = bank_percentage
        self.withdrawal_loss = withdrawal_loss
        if alpha != None: 
            self.alpha = np.array(alpha, dtype=np.float64, ndmin=1)
        else: 
            self.alpha = None   

        # run checks 
        self._checks() 
        
        # define state and decision spaces 
        self._state_space() 
        self._decision_space() 
        self._inflow_space() 
        self._check_post()
        
        # program prints 
        self.nprints = 0 
        
        # it is now initialized
        self.initialized = True 
        
    def _checks(self):
        # user info 
        assert len(self.users) == len(self.abbrev) and \
               len(self.funcs) == len(self.users) and \
               self.stepsizes.shape[1] == len(self.users), \
               'The number of users, abbreviations, demand functions, and stepsizes for users (step_users) must be equivalent!'
        for ubound in self.upperbounds: 
            assert len(ubound) == len(self.users), \
                   'Each upperbound definition in "upperbounds" must have the same number of elements as the number of users!'
            
        # period info 
        assert len(self.periods) == self.upperbounds.shape[0] and \
               len(self.periods) == self.stepsizes.shape[0] and \
               len(self.periods) == len(self.spill_upper) and \
               len(self.periods) == len(self.spill_step), \
               'The number of elements in "upperbounds" and "stepsizes" and "spill_upper" and "spill_step" must match the number of elements in "periods"!'
        
        # state info
        assert len(self.storage_capacity) == 1, \
            'Input parameter "storage_capacity" must be a scalar.'

        # allotment info
        if self.alpha == None: 
            assert self.scen_type == ScenarioType.FreeMarket or \
                   self.n_type == ScenarioType.FreeMarketBanking, \
                   'alpha must be defined  for any Prior Appropriation scenario.'
        else: 
            assert self.alpha.sum() == 1, \
                'The allotted percentages given in "alpha" must sum to 1!'
            assert len(self.alpha) == len(self.users), \
                'The allotted percentages given in "alpha" must have the same number of elements as the number of users!'

    def _check_post(self):
        assert len(self.decision_stepsizes) == self.ndecisions, \
            'The number of stepsizes in decision_stepsizes needs to match the number of decisions in ndecisions'
        assert len(self.decision_names) == self.ndecisions, \
            'The number of names in decision_names needs to match the number of decisions in ndecisions'
        assert len(self.decision_upperbounds) == self.ndecisions, \
            'The number of upper bounds in decision_upperbounds needs to match the number of decisions in ndecisions'
        assert len(self.state_stepsizes) == self.nstates, \
            'The number of stepsizes in state_stepsizes needs to match the number of states in nstates'
        assert len(self.state_names) == self.nstates, \
            'The number of names in state_names needs to match the number of states in nstates'
        assert len(self.state_upperbounds) == self.nstates, \
            'The number of upper bounds in state_upperbounds needs to match the number of states in nstates'
        
    def _state_space(self):
        print 'self.scen_type', self.scen_type
        # state space  
        self.nstates = \
            {ScenarioType.FreeMarket:          1,
             ScenarioType.FreeMarketBanking:   1 + self.nusers,
             ScenarioType.PriorApprop:         1,
             ScenarioType.PriorAppropBanking:  1 + self.nusers}[self.scen_type]
        self.state_stepsizes = \
            {ScenarioType.FreeMarket:          np.array(self.stepsizes.min(), ndmin=1), 
             ScenarioType.FreeMarketBanking:   np.hstack((self.stepsizes.min(), self.stepsizes.min(axis=0))),
             ScenarioType.PriorApprop:         np.array(self.stepsizes.min(), ndmin=1), 
             ScenarioType.PriorAppropBanking:  np.hstack((self.stepsizes.min(), self.stepsizes.min(axis=0)))}[self.scen_type]
        leftover = self.storage_capacity % self.state_stepsizes[0]
        if leftover != 0.0: 
            self.storage_capacity += self.state_stepsizes[0] - leftover
        self.minstor = np.array(0.0, ndmin=1)
        self.maxstor = np.array(self.storage_capacity, ndmin=1)
        self.state_upperbounds = \
            {ScenarioType.FreeMarket:          self.storage_capacity, 
             ScenarioType.FreeMarketBanking:   np.hstack((self.storage_capacity, self.bank_percentage * self.upperbounds.sum(axis=0))),
             ScenarioType.PriorApprop:         self.storage_capacity, 
             ScenarioType.PriorAppropBanking:  np.hstack((self.storage_capacity, self.bank_percentage * self.upperbounds.sum(axis=0)))}[self.scen_type]
        self.state_names = \
            {ScenarioType.FreeMarket:          ['Storage'], 
             ScenarioType.FreeMarketBanking:   ['Storage'] + ['Bank_%s' % u for u in self.users],
             ScenarioType.PriorApprop:         ['Storage'], 
             ScenarioType.PriorAppropBanking:  ['Storage'] + ['Bank_%s' % u for u in self.users]}[self.scen_type]
        self.state_indices = \
            {ScenarioType.FreeMarket:          (0, None), 
             ScenarioType.FreeMarketBanking:   (0, range(1, self.nusers + 1)),
             ScenarioType.PriorApprop:         (0, None), 
             ScenarioType.PriorAppropBanking:  (0, range(1, self.nusers + 1))}[self.scen_type]
            
    def _decision_space(self):    
        # decision space
        s1 = self.stepsizes.flatten()          # nperiods * nusers
        s2 = self.stepsizes.min(axis=1)        # nperiods
        s3 = self.spill_step                   # nperiods
        u1 = self.upperbounds.flatten()        # nperiods * nusers 
        u2 = self.upperbounds.sum(axis=1)      # nperiods 
        u3 = self.spill_upper                  # nperiods
        u4 = self.bank_percentage * u1         # nperiods * nusers 
        nt = self.nperiods * self.nusers 
        nper = self.nperiods
         
        self.ndecisions = \
            {ScenarioType.FreeMarket:          self.nperiods * (self.nusers + 1), 
             ScenarioType.FreeMarketBanking:   self.nperiods * (self.nusers*3 + 1),
             ScenarioType.PriorApprop:         self.nperiods * 2,
             ScenarioType.PriorAppropBanking:  self.nperiods * (self.nusers*2 + 2)}[self.scen_type] 
        self.decision_stepsizes = \
            {ScenarioType.FreeMarket:          np.hstack((s1, s3)),  
             ScenarioType.FreeMarketBanking:   np.hstack((s1, s1, s1, s3)),
             ScenarioType.PriorApprop:         np.hstack((s2, s3)), 
             ScenarioType.PriorAppropBanking:  np.hstack((s2, s1, s1, s3))}[self.scen_type]
        self.decision_upperbounds = \
            {ScenarioType.FreeMarket:          np.hstack((u1, u3)),  
             ScenarioType.FreeMarketBanking:   np.hstack((u1, u4, u4, u3)),
             ScenarioType.PriorApprop:         np.hstack((u2, u3)), 
             ScenarioType.PriorAppropBanking:  np.hstack((u2, u4, u4, u3))}[self.scen_type]
        self.decision_names = \
            {ScenarioType.FreeMarket:          ['Release_%s_%s' % (u,p) for p, u in itertools.product(self.periods, self.users)] + 
                                               ['Spill_%s' % p for p in self.periods], 
             ScenarioType.FreeMarketBanking:   ['Release_%s_%s' % (u,p) for p, u in itertools.product(self.periods, self.users)] + 
                                               ['Save_%s_%s' % (u,p) for p, u in itertools.product(self.periods, self.users)] + 
                                               ['Withdraw_%s_%s' % (u,p) for p, u in itertools.product(self.periods, self.users)] + 
                                               ['Spill_%s' % p for p in self.periods],
             ScenarioType.PriorApprop:         ['TotalRelease_%s' % p for p in self.periods] + 
                                               ['Spill_%s' % p for p in self.periods], 
             ScenarioType.PriorAppropBanking:  ['TotalRelease_%s' % p for p in self.periods] +  
                                               ['Save_%s_%s' % (u,p) for p, u in itertools.product(self.periods, self.users)] + 
                                               ['Withdraw_%s_%s' % (u,p) for p, u in itertools.product(self.periods, self.users)] + 
                                               ['Spill_%s' % p for p in self.periods]}[self.scen_type]
        self.decision_indices = \
            {ScenarioType.FreeMarket:          (range(nt), None, None, range(nt,nt+nper)), 
             ScenarioType.FreeMarketBanking:   (range(nt), range(nt,2*nt), range(2*nt, 3*nt), range(3*nt, 3*nt + nper)),
             ScenarioType.PriorApprop:         (range(nper), None, None, range(nper,2*nper)), 
             ScenarioType.PriorAppropBanking:  (range(nper), range(nper,nper+nt), range(nper+nt,nper+2*nt), range(nper+2*nt, 2*nper+2*nt))}[self.scen_type]

    def _inflow_space(self): 
        # check file existence and read file 
        if not os.path.exists(self.inflowfile):
            raise Exception(self.inflowfile + ' must exist in order to run dp problem.')
        allinflow = np.loadtxt(self.inflowfile, dtype=np.float64, delimiter='\t', skiprows=1)
        
        # check timesteps
        tsteps = list(set(allinflow[:, 0]))
        tsteps.sort()
        assert len(tsteps) == len(self.periods), 'The number of timesteps in the inflowprob.txt file must match the number of timesteps specified in the problem definition'
        
        # build the inflow list 
        self.inflow = []
        self.inflowprob = [] 
        for tstep in tsteps:
            rows = (allinflow[:, 0] == tstep)
            currinf = allinflow[rows][:, 1]
            self.inflow.append(currinf.reshape((-1,1)))
            self.inflowprob.append(allinflow[rows][:, 2])
        self.ninflows = np.mean([len(i) for i in self.inflow])
            
    def get_period(self, t):
        return int(t % self.nperiods)
            
    def get_grid(self, state):
        '''
        Retrieves the stategrid or decisiongrid for this problem
        '''
        steps = self.state_stepsizes if state else self.decision_stepsizes
        upper = self.state_upperbounds if state else self.decision_upperbounds
        lower = np.zeros(len(upper))
        return getgrid(lower, upper, steps)
    
    def get_state_indices(self, bank):
        '''
        Retrieves the indices within the state variable for a particular class 
        of data
        '''
        return self.state_indices[int(bank)]

    def get_decision_indices(self, type):
        '''
        Retrieves the indices within the decision variable for a particular 
        class of data and period.
        
        @param type:        The class of data that is of interest. Can be either 
                            DecisionType.Release or DecisionType.Spill, and 
                            in the case of banking this can also be 
                            DecisionType.Save or DecisionType.Withdraw 
        '''
        # get the range of indices for the specified class (type) 
        allind = self.decision_indices[type]
        
        # get the length of each period's chunk (should be an integer) 
        chunklen = int(len(allind) / self.nperiods)
        
        # get the range of indices for the specified period (in addition to the 
        # specified class) 
        return [allind[p*chunklen:(p+1)*chunklen] for p in range(self.nperiods)] 

    def _fxns_FreeMarket(self):
        
        # decision indices 
        qi = self.get_decision_indices(DecisionType.Release)
        qspill = self.get_decision_indices(DecisionType.Spill)
        
        def statefxn(s, x, t, inflowval):
            return s + inflowval - x.sum()
    
        def fixdec(x):
            x[qi] = np.minimum(x[qi], self.demfxn.opt_decision)
             
        def fixexcess(x, excess):
            qold = x.sum()
            def get_value(q): 
                return -self.demfxn.total_value(q[qi])
            def get_deriv(q):
                return np.hstack((-self.demfxn.marg_value(q[qi]),0.0))
            def mass_bal(q): 
                return q.sum() - qold - excess
            def mass_deriv(q): 
                return (1, 1, 1)
            x[:] = minimize(get_value, x, 
                            jac=get_deriv, 
                            method='SLSQP', 
                            constraints=({'type':'eq','fun':mass_bal,'jac':mass_deriv}), 
                            bounds=([(0.0, self.demfxn.opt_decision[i]) for i in range(len(self.demfxn.opt_decision))] + 
                                    [(0.0, 1e+08)]),
                            options={'disp': False}).x
            print x.sum() - qold - excess
         
        def fixshortage(x, shortage): 
            fixexcess(x, -shortage)
                 
        def simfxn(s, x, t, inflowval):
            # calculate next state
            fixdec(x) 
            snext = statefxn(s, x, t, inflowval)
     
            if (snext > self.maxstor).any():
                # TOO MUCH WATER - SPILL IF NO OTHER USERS WANT IT 
                excess = snext-self.maxstor  
                fixexcess(x, excess)
                if self.nprints < self.maxnumerrprints: 
                    print 'EXCESS WATER AT TIME %d: %s' % (t, str(excess))
                    print '  Setting spill to %s' % str(x[qspill])
                    print '  Inflow: %s' % str(inflowval)
                    print '  Decisions: %s' % str(x)
                    self.nprints += 1
                elif self.nprints == self.maxnumerrprints: 
                    print 'EXCEEDED STORAGE MORE THAN %d TIMES, NOT PRINTING ANY MORE OUTPUT...' % self.maxnumerrprints
                    self.nprints += 1
                  
                return (self.maxstor, x)
            elif (snext < self.minstor).any():
                # NOT ENOUGH WATER - DECREASE DISCHARGE TO WATER USERS
                # Currently, only the scheme where everybody loses out on the water 
                #   by a certain percentage
                shortage = self.minstor-snext
                fixshortage(x, shortage)
                if self.nprints < self.maxnumerrprints: 
                    print 'SHORTAGE OF WATER AT TIME %d: %s' % (t, str(shortage)) 
                    print '  Setting discharge to: %s ' % str(x[qi])
                    print '  Inflow: %s' % str(inflowval)
                    print '  Decisions: %s' % str(x)
                    self.nprints += 1
                elif self.nprints == self.maxnumerrprints: 
                    print 'DRAINED STORAGE MORE THAN %d TIMES, NOT PRINTING ANY MORE OUTPUT...' % self.maxnumerrprints
                    self.nprints += 1
                return (self.minstor, x)
            else:
                # WE'RE OK RETURN. HOPEFULLY WE'RE HERE MOST OFTEN
                return (snext, x)    

        def valuefxn(s, snext, x, t, inflowval):
            if (snext < self.minstor).any():   # EMPTIED STORAGE TOO MUCH
                return -np.inf
            elif (snext > self.maxstor).any(): # FILLED RESERVOIR TOO MUCH
                return -np.inf
            return self.demfxn.total_value(x[qi])

        return statefxn, simfxn, valuefxn
    
    def _fxns_PriorApprop(self): 
        
        # decision indices 
        qi = self.get_decision_indices(DecisionType.Release)
        qspill = self.get_decision_indices(DecisionType.Spill)

        def statefxn(s, x, t, inflowval):
            return s + inflowval - x.sum()
        
        def fixdec(x):
            x[qi] = min(x[qi].sum(), (self.demfxn.opt_decision / self.alpha).min())

        def fixexcess(x, excess):
            qold = x.sum()
            x[qi] = min(qold + excess, (self.demfxn.opt_decision / self.alpha).min())
            x[qspill] += qold + excess - x.sum()
         
        def fixshortage(x, shortage): 
            x[qi] = min(x.sum() - shortage, (self.demfxn.opt_decision / self.alpha).min())

        def simfxn(s, x, t, inflowval):
            # calculate next state
            fixdec(x) 
            snext = statefxn(s, x, t, inflowval)
     
            if (snext > self.maxstor).any():
                # TOO MUCH WATER - SPILL IF NO OTHER USERS WANT IT 
                excess = snext-self.maxstor  
                fixexcess(x, excess)
                if self.nprints < self.maxnumerrprints: 
                    print 'EXCESS WATER AT TIME %d: %s' % (t, str(excess))
                    print '  Setting spill to %s' % str(x[qspill])
                    print '  Inflow: %s' % str(inflowval)
                    print '  Decisions: %s' % str(x)
                    self.nprints += 1
                elif self.nprints == self.maxnumerrprints: 
                    print 'EXCEEDED STORAGE MORE THAN %d TIMES, NOT PRINTING ANY MORE OUTPUT...' % self.maxnumerrprints
                    self.nprints += 1
                  
                return (self.maxstor, x)
            elif (snext < self.minstor).any():
                # NOT ENOUGH WATER - DECREASE DISCHARGE TO WATER USERS
                # Currently, only the scheme where everybody loses out on the water 
                #   by a certain percentage
                shortage = self.minstor-snext
                fixshortage(x, shortage)
                if self.nprints < self.maxnumerrprints: 
                    print 'SHORTAGE OF WATER AT TIME %d: %s' % (t, str(shortage)) 
                    print '  Setting discharge to: %s ' % str(x[qi])
                    print '  Inflow: %s' % str(inflowval)
                    print '  Decisions: %s' % str(x)
                    self.nprints += 1
                elif self.nprints == self.maxnumerrprints: 
                    print 'DRAINED STORAGE MORE THAN %d TIMES, NOT PRINTING ANY MORE OUTPUT...' % self.maxnumerrprints
                    self.nprints += 1
                return (self.minstor, x)
            else:
                # WE'RE OK RETURN. HOPEFULLY WE'RE HERE MOST OFTEN
                return (snext, x)    
        
        def valuefxn(s, snext, x, t, inflowval):
            if (snext < self.minstor).any():   # EMPTIED STORAGE TOO MUCH
                return -np.inf
            elif (snext > self.maxstor).any(): # FILLED RESERVOIR TOO MUCH
                return -np.inf
            return self.demfxn.total_value(x[qi] * self.alpha)

        return statefxn, simfxn, valuefxn 
    
    def _fxns_PriorAppropBanking(self):
        # decision indices 
        qi = self.get_decision_indices(DecisionType.Release)
        qsavei = self.get_decision_indices(DecisionType.Save)
        qwdrwi = self.get_decision_indices(DecisionType.Withdraw)
        qspill = self.get_decision_indices(DecisionType.Spill)

        # state indices 
        stori = self.get_state_indices(False)
        banki = self.get_state_indices(True)

        # next state 
        nextstate = np.zeros(self.nstates)

        def statefxn(s, x, t, inflowval, retall=False):
            # period 
            p = self.get_period(t) 
            
            # decision calculations 
            xallot = self.alpha * x[qi[p]]
            xsave = 1 * x[qsavei[p]] # make a new array 
            xwdrw = 1 * x[qwdrwi[p]]
            xuse = xallot - xsave + (1 - withdrawal_loss) * xwdrw
            
            # state calculations 
            nextstate[stori] = s[stori] + inflowval - xuse.sum()
            nextstate[banki] = s[banki] + xsave - xwdrw
            
            # return 
            if retall: 
                return (nextstate, xuse, xallot, xsave, xwdrw)
            else: 
                return nextstate
    
        def simfxn(s, x, t, inflowval):
            # fix state at physical reservoir and bank
            x = fix_decision(x) 
            snext, xuse, xallot, xsave, xwdrw = statefxn(s, x, t, inflowval, retall=True)
             
            # fix anybody that attempts to save more than alloted... 
            savedtoomuch = (xsave > xallot)
            if savedtoomuch.any(): 
                x[qsavei][savedtoomuch] = xallot[savedtoomuch]
                print 'The following users attempted to save more than their allotment:\n', usernames[savedtoomuch]
             
            # storage problems
            if (snext[stori] > self.maxstor[stori]).any():
                # TOO MUCH WATER - CAN'T SAVE AS MUCH AND, IF NEED BE, SEND MORE 
                #   WATER OUT THE SPILLWAY
                excess = snext[stori] - self.maxstor[stori]
                totxsave = xsave.sum()
                if excess < totxsave:
                    # REDUCE AMOUNT FOR SAVINGS
                    p = 1. - excess / totxsave
                    x[qsavei] = p * xsave
                    if self.nprints < self.maxnumerrprints: 
                        print 'EXCESS STORAGE AT TIME %d: %s' % (t, str(excess))
                        print '  Reduced savings by %f\%' % p
                        self.nprints += 1
                    elif self.nprints == self.maxnumerrprints: 
                        print 'EXCEEDED STORAGE MORE THAN %d TIMES, NOT PRINTING ANY MORE OUTPUT...' % self.maxnumerrprints
                        self.nprints += 1
                else: 
                    # SET SAVINGS TO ZERO AND SPILL MORE
                    x[qsavei] = 0
                    x[qspill] += excess - totxsave
                    if self.nprints < self.maxnumerrprints: 
                        print 'EXCESS STORAGE AT TIME %d: %s' % (t, str(excess))
                        print '  Require spill of %f due to capacity' % x[qspill]
                        self.nprints += 1
                    elif self.nprints == self.maxnumerrprints: 
                        print 'EXCEEDED STORAGE MORE THAN %d TIMES, NOT PRINTING ANY MORE OUTPUT...' % self.maxnumerrprints
                        self.nprints += 1
                snext[stori] = self.maxstor[stori]
                snext[banki] = snext[banki] - xsave + x[qsavei]
            elif (snext[stori] < self.minstor[stori]).any():
                # NOT ENOUGH WATER - DECREASE DISCHARGE TO WATER USERS
                # Currently, only the scheme where everybody loses out on the  
                #   water by a certain percentage
                # Start with decreasing alloted discharge, then decrease banked 
                #   withdrawals.
                lacking = self.minstor[stori] - snext[stori]
                totxallot = xallot.sum()
                if lacking < totxallot: 
                    # REDUCE DISCHARGE FROM CURRENT ALLOTMENT
                    # x[qi] is the total discharged allotment across all users,
                    #   so we just need to reduce
                    p = 1. - lacking / totxallot
                    x[qi] = p * x[qi]
                    if self.nprints < self.maxnumerrprints: 
                        print 'DRAINED STORAGE AT TIME %d: %s' % (t, str(lacking))
                        print '  Decreasing user allotment by percentage:',p     
                        self.nprints += 1
                    elif self.nprints == self.maxnumerrprints: 
                        print 'DRAINED STORAGE MORE THAN %d TIMES, NOT PRINTING ANY MORE OUTPUT...' % self.maxnumerrprints
                        self.nprints += 1
                else: 
                    # SET ALLOTED DISCHARGE TO ZERO AND REDUCE BANK WITHDRAWALS
                    p = 1 - (lacking - totxallot) / ((1 - withdrawal_loss) * xwdrw).sum()
                    x[qi] = 0                     
                    x[qwdrwi] = p * xwdrw
                    if self.nprints < self.maxnumerrprints: 
                        print 'DRAINED STORAGE AT TIME %d: %s' % (t, str(lacking))
                        print '  Decreasing user allotment by:', totxallot
                        print '  Decreasing bank withdrawals by percentage:', p
                        self.nprints += 1
                    elif self.nprints == self.maxnumerrprints: 
                        print 'DRAINED STORAGE MORE THAN %d TIMES, NOT PRINTING ANY MORE OUTPUT...' % self.maxnumerrprints
                        self.nprints += 1
                snext[stori] = self.minstor[stori] 
                snext[banki] = snext[banki] - xwdrw + x[qwdrwi]
                 
            # return 
            # check one last time 
            if (snext < self.minstor).any() or (snext > self.maxstor).any():
                excstr = 'Storage was again not satisfied in simfxn!!!\n' + \
                         '  t       : %s\n' % str(t) + \
                         '  s       : %s\n' % str(s) + \
                         '  x       : %s\n' % str(x) + \
                         '  inflow  : %s\n' % str(inflowval) + \
                         '  snext   : %s\n' % str(snext) + \
                         '  minstor : %s\n' % str(self.minstor) + \
                         '  maxstor : %s\n' % str(self.maxstor)
                raise Exception(excstr)            
            return (snext, x)
 
        def valuefxn(s, snext, x, t, inflowval):
            # get the amount of water used 
            _, xuse, xallot, xsave, xwdrw = \
                statefxn(s, x, t, inflowval, retall=True)
            
            # ensure feasibility  
            if (xsave > xallot).any(): # INFEASIBLE DECISION FOR WATER USE
                return -np.inf
            if (snext < self.minstor).any():   # EMPTIED STORAGE TOO MUCH
                return -np.Inf
            elif (snext > self.maxstor).any(): # FILLED RESERVOIR TOO MUCH
                return -np.Inf
            
            # return the calculated value
            return self.demfxn.total_value(xuse)
        
        return statefxn, simfxn, valuefxn
    
    def get_transitionP(self):
        def transProb(s, nexts, x, t, rndvar_i): # p, x, t, w):
            return self.inflowprob[self.get_period(t)][rndvar_i]
        return transProb
        
    def get_rndvar_space(self):
        def rndvar_space(s, t):
            return self.inflow[self.get_period(t)]
        return rndvar_space
    
    def get_fxns(self):
        pr = getattr(self, '_fxns_' + ScenarioType.get_name(self.scen_type))
        return pr() 
        
    def load(self):
        self.results = dp.load_solution(self.outdir, self.scen_name)
        return self.results 
    
    def save(self, results=None):
        if results == None: 
            results = self.results
        dp.save_solution(self.outdir, self.scen_name, results)

    def initialvars(self, init_state, T):
        # set up model parameters 
        state0 = np.array(init_state, ndmin=1, dtype=np.float64)
        t = list(np.linspace(0, T, T + 1))
        
        # maximize or minimize? 
        ismax = True
    
        # iteration tolerance
        tolerance = 1e-03
       
        # grids 
        decisiongrid = self.get_grid(state=False)
        stategrid = self.get_grid(state=True)
        transProb = self.get_transitionP()
        rndvar_space = self.get_rndvar_space()
        statefxn, simfxn, valuefxn = self.get_fxns()
        return state0, t, ismax, tolerance, decisiongrid, stategrid,transProb, \
               rndvar_space, statefxn, simfxn, valuefxn

    def solve(self,
            # time horizon
            infhorizon=False, T=50,
            # optimization options
            acceptablerisk=0.1, discount=1, seed=456, 
            # state variables
            init_state=250000, 
            # program options  
            verbose=False, nthreads=1):  
        '''
        Solves a dynamic programming setup of the reservoir problem for a single 
        reservoir
        '''
        # number of print statements
         
        # calculate the speed of solution 
        starttime = time.time()

        # initialize variables     
        state0,t,ismax,tolerance,decisiongrid,stategrid,transProb,rndvar_space,\
            statefxn,simfxn,valuefxn = self.initialvars(init_state, T)
             
        ####################### setup notification ################################# 
    
        # show problem set up
        print 'Decision dimensions:          %d' % self.ndecisions
        print 'State dimensions:             %d' % self.nstates
        print 'Decision space realizations:  %d' % decisiongrid.ncombos
        print 'State space realizations:     %d' % stategrid.ncombos
        print 'Average inflow realizations:  %d' % self.ninflows
        print 'Evaluations at each state:    %d' % (decisiongrid.ncombos * self.ninflows) 
        print 'Evaluations at each stage t:  %d' % (stategrid.ncombos * decisiongrid.ncombos * self.ninflows)
        print 'Total number of evaluations:  %d' % (T * stategrid.ncombos * decisiongrid.ncombos * self.ninflows)
        print ''
    
        # show amount of time
        endtime = time.time()
        print 'Took %g seconds to initialize.' % (endtime - starttime)
        starttime = endtime
    
    
        ####################### solve the problem ################################## 
        np.random.seed(seed) # seed all random calculations
        if infhorizon: 
            self.results = \
                dp.solve_infinitetime(valuefxn, statefxn, stategrid, decisiongrid, 
                                      tolerance=tolerance, rndvar_space=rndvar_space, 
                                      transitionP=transProb, ismax=ismax, 
                                      discount=discount, verbose=verbose, 
                                      acceptablerisk=acceptablerisk,
                                      nthreads=nthreads)
        else:
            self.results = \
                dp.solve_finitetime(valuefxn, statefxn, stategrid, decisiongrid, t,
                                      rndvar_space=rndvar_space, 
                                      transitionP=transProb, ismax=ismax, 
                                      discount=discount, verbose=verbose, 
                                      acceptablerisk=acceptablerisk,
                                      nthreads=nthreads)
    
        # show amount of time
        endtime = time.time()
        print 'Took %g seconds to solve.' % (endtime - starttime)
        
        self.save()
        print 'Present Value =', self.results['opt_values'][0]
        return self.results 
    
    def trace(self, T, nscenarios=100, seed=456, init_state=250000, 
              verbose=False, discount=1):
        
        # start timer
        starttime = time.time()
         
        # initialize variables 
        state0,t,ismax,tolerance,decisiongrid,stategrid,transProb,rndvar_space,\
            statefxn,simfxn,valuefxn = self.initialvars(init_state, T)
        
        # seed all random calculations for tracing
        np.random.seed(seed) 
        
        # trace optimal results
        trace_results = dp.trace_optimal_paths(stategrid, decisiongrid, t, state0,
                        self.results, ismax=ismax, verbose=verbose,
                        rndvar_space=rndvar_space, transitionP=transProb, 
                        valuefxn=valuefxn, simfxn=simfxn, 
                        nscenarios=nscenarios, discount=discount,
                        errorOnInfeasible=False, snapvals=False)            
    
        # show amount of time
        endtime = time.time()
        print 'Took %g seconds to trace results.' % (endtime - starttime)
        starttime = endtime   
    
        # save main results
        self.save(trace_results)
        
        # add to results dictionary 
        for key in trace_results: 
            self.results[key] = trace_results[key]
    
        # save appropriated amounts
        decisions = np.array(trace_results['trace_decisions']) 
        if self.scen_type == ScenarioType.PriorApprop or \
           self.scen_type == ScenarioType.PriorAppropBanking:
            
            # indices for total release 
            qi = np.asarray(self.get_decision_indices(DecisionType.Release)).reshape((-1))
        
            # appropriated amount to each user  
            appramt = decisions[:,:,qi] * self.alpha
            
            # save the appropriated amounts to file 
            outname = os.path.join(self.outdir, self.scen_name) + 'trace_appramt.npy' 
            np.save(outname, appramt) 
            
            # get marginal values 
            mvals = np.zeros((decisions.shape[0], decisions.shape[1], self.nusers))
            for t in range(decisions.shape[0]): 
                for i in range(decisions.shape[1]): 
                    d = decisions[t,i,qi] #.reshape((-1))
                    mvals[t, i, :] = self.demfxn.marg_value(d * self.alpha)
        elif self.scen_type == ScenarioType.FreeMarket:
            
            # indices for water delivery to each user
            qi = np.asarray(self.get_decision_indices(DecisionType.Release)).reshape((-1))
            
            # get marginal values  
            mvals = np.zeros((decisions.shape[0], decisions.shape[1], len(qi)))
            for t in range(decisions.shape[0]): 
                for i in range(decisions.shape[1]): 
                    mvals[t, i, :] = self.demfxn.marg_value(decisions[t,i,qi].reshape((-1)))
        else: 
            raise Exception('Marginal value calculation has not been implemented for scentype %s.' % getkey(ScenarioType.__dict__, scentype))
        
        # save marginal values 
        outname = os.path.join(self.outdir, self.scen_name) + 'trace_margvalues.npy'
        np.save(outname, mvals)

        # time 
        endtime = time.time()
        print 'Took %g seconds to save traced results.' % (endtime - starttime)
        
        return self.results
    
    def plot(self, verbose=False, showplots=False):
        '''
        Plots trace plots and histograms of present values of optimization and 
        subsequent simulation of C-BT system
        '''
        try: 
            import matplotlib.pyplot as plt
        except:
            print 'Not able to open matplotlib, not saving any plots.'
            return
    
        starttime = time.time()

        # close all figures 
        plt.close('all')
    
        # output prefix 
        if not os.path.exists(self.outdir):
            os.makedirs(self.outdir)
        totprefix = os.path.join(self.outdir, self.scen_name)
    
        # load the desired data 
        results = dp.load_solution(self.outdir, self.scen_name)
        decisions = results['trace_decisions'] 
        states = results['trace_states'] 
        values = results['trace_values']
        incrvalues = results['trace_incrvalues']
        margvalues = results['trace_margvalues']
        rndvars = None 
        if 'trace_rndvars' in results: 
            rndvars = results['trace_rndvars']
        appramt = None
        if 'trace_appramt' in results: 
            appramt = results['trace_appramt']
        maxerrors = None 
        if 'maxerrors' in results: 
            maxerrors = results['maxerrors'] 
    
        # fix variables
        states = np.array([s.reshape((-1,)) for s in states])
    
        # set variables
        T = values.shape[0] - 1
        t = list(np.linspace(0, T, T + 1))
        flh='Time increases down the rows; inflow realizations vary across columns.'
        
        ############################################################################
        ####################### define functions ###################################
        ############################################################################
    
        class plot_type: 
            # the type to plot 
            State = 0
            Decision = 1
            Values = 2
            IncrValues = 3
            MaxErrors = 4 
            MargValues = 5
            ApprAmt = 6
            RndVar = 7
            
            @staticmethod
            def get_name(ptype): 
                return getkey(plot_type.__dict__, ptype)
        
        def get_ylabel(ptype, multi_i=0):
            # get the ylabel for the plot 
            if ptype==plot_type.Decision: 
                return 'Decision path for %s (acre-feet/year)' % self.decision_names[multi_i]
            elif ptype==plot_type.State: 
                return 'State path for %s (acre-feet)' % self.state_names[multi_i]
            elif ptype==plot_type.Values:
                return 'Total discounted value path ($)'
            elif ptype==plot_type.IncrValues: 
                return 'Current value path ($)'
            elif ptype==plot_type.MargValues: 
                return 'Marginal value path for %s ($/AF)' % self.users[multi_i]
            elif ptype==plot_type.ApprAmt: 
                return 'Appropriated amount to %s (acre-feet/year)' % self.users[multi_i]
            elif ptype==plot_type.MaxErrors:
                return 'Maximum error during infinite horizon iterations'
            elif ptype==plot_type.RndVar:
                return 'Random variable realization'
            else: 
                raise Exception('ptype %s is not implemented.' % str(ptype))
        
        def get_fig_suffix(ptype, multi_i=0): 
            # get the suffix for the figure output file name
            if ptype==plot_type.Decision: 
                return '_' + self.decision_names[multi_i]
            elif ptype==plot_type.MargValues:
                return '_' + self.users[multi_i]
            elif ptype==plot_type.ApprAmt:
                return '_' + self.users[multi_i]
            elif ptype==plot_type.State:
                return '_' + self.state_names[multi_i]
            else: 
                return ''
    
        def format_data(vals): 
            # determine if there is multiple dimensions of the variable
            multidim =(len(vals.shape) > 2)  
            if multidim: 
                vals = [np.array([v.reshape((-1)) for v in vals[:, :, i]]) for i in range(vals.shape[2])]
                means = [v.mean(axis=1) for v in vals]
            else: 
                vals = np.array([v.reshape((-1)) for v in vals])
                means = vals.mean(axis=1)
            return vals, means, multidim 
        
        def plot_single(vals, ptype):
            # plot a single type of output (state, value, decision, error, incrvalue, or margvalue)
            if vals == None: 
                return 
            vals, means, multidim = format_data(vals)
            
            # plot multiple dimensions in different figures 
            if multidim: 
                for i in range(len(vals)): 
                    plt.figure() 
                    lent = len(vals[i])
                    timearray = np.array(t[:lent]).reshape((-1,1))
                    plt.plot(timearray, vals[i], 'k-', alpha=0.2)
                    plt.plot(timearray, means[i], 'r-', linewidth=3)
                    plt.xlabel('Time') 
                    plt.ylabel(get_ylabel(ptype, i))
                    outname = totprefix + plot_type.get_name(ptype) + \
                                get_fig_suffix(ptype, i)
                    try: 
                        np.savetxt(outname + '.csv', vals[i], delimiter=',', header=flh)
                    except: 
                        pass
                    plt.savefig(outname + '.png', dpi=300)
            else: 
                plt.figure()
                lent = len(vals)
                timearray = np.array(t[:lent]).reshape((-1,1))
                plt.plot(timearray, vals, 'k-', alpha=0.2)
                plt.plot(timearray, means, 'r-', linewidth=3)
                plt.xlabel('Time') 
                plt.ylabel(get_ylabel(ptype))
                outname = totprefix + plot_type.get_name(ptype) 
                try: 
                    np.savetxt(outname + '.csv', vals, delimiter=',', header=flh)
                except: 
                    pass
                plt.savefig(outname + '.png', dpi=300)
    
        def plot_init_values(vals, ptype=plot_type.Values):
            # plot the initial simulated values as a histogram  
            vals, means, multidim = format_data(vals)
            plt.figure()
            init_vals = vals[0][0, :]
            plt.hist(init_vals, color='k', alpha=0.4, normed=True, histtype='stepfilled')
            plt.xlabel(get_ylabel(ptype))
            plt.ylabel('Density')
            plt.savefig(totprefix + 'Init' + plot_type.get_name(ptype) + \
                        get_fig_suffix(ptype) + '.png', dpi=300)
    
            # summarize and save present values
            percs = tuple([init_vals.mean(), init_vals.std()] + np.percentile(list(init_vals), [0, 2.5, 25, 50, 75, 97.5, 100]))
            headers = ['Mean', 'StdDev', '2.5th', 'Min', '25th', 'Median', '75th', '97.5th', 'Max']
            
            # print 
            strpercs = [headers[i].ljust(7) + ' = ' + str(percs[i]) for i in range(len(percs))]
            sep = '\n    '
            print '\nPresent Values:%s%s\n' % (sep, string.join(strpercs, ',' + sep))
            
            # write to file
            with open(totprefix + 'PresentValues.csv', 'w') as f: 
                wr = csv.writer(f) 
                wr.writerow(headers)
                wr.writerow(percs)
                
        ############################################################################
        ####################### plot each output variables #########################
        ############################################################################
        
        # plot each of the traced solutions 
        plot_single(decisions, plot_type.Decision)
        plot_single(states, plot_type.State)
        plot_single(values, plot_type.Values)
        plot_single(incrvalues, plot_type.IncrValues)
        plot_single(margvalues, plot_type.MargValues)
        plot_single(appramt, plot_type.ApprAmt)
        plot_single(maxerrors, plot_type.MaxErrors)
        plot_single(rndvars, plot_type.RndVar)
        plot_init_values(values)
        
        # show plots if desired
        if showplots: 
            plt.ion()
            plt.show() 

        endtime = time.time()
        print 'Took %g seconds to plot results.' % (endtime - starttime)
    
        return self.results
     
         