'''
Created on Nov 11, 2013

@author: Andre Dozier
@copyright: 2014 Andre Dozier  
@summary:   The purpose of this code is to perform generic dynamic programming 
            given simple user-specified functions. 

            This file was created in partial requirements for AREC 540 (Natural 
            Resources Economics) at Colorado State University in Fall 2013 by 
            a full-time CSU Ph.D. student Andre Dozier.  
'''

import numpy as np 
import combinatorics
from copy import copy
from numpy.random import random_sample 
import sys, time, glob, os
from ThreadPool import ThreadPool 

numofprints = 0
maxnumofprints = 50
numoflargeevals = 0
maxnumoflargeevals = 4
maxsecondsperstage = 900. / 50.
evaltimes = None

def ensurelist(value):
    # ensure that the value is a list
    if isinstance(value, list): 
        return value 
    elif isinstance(value, np.ndarray): 
        return list(value)
    else: 
        return [value]
    
def ensurearray(value, isfloat):
    # ensure that the value is converted properly to an array
    dtype = np.float64 if isfloat else np.int32 
    if isinstance(value, list): 
        return np.array(value, dtype=dtype, ndmin=1)
    elif not isinstance(value, np.ndarray): 
        return np.array(ensurelist(value), dtype=dtype, ndmin=1)
    else: 
        return np.array(value, dtype=dtype, ndmin=1)
        
class LinGrid(object):
    '''
    A class that defines a linear grid based on minimum, maximum, and 
    number of values for any number (finite) of dimensions
    
    For example: 
        l = LinGrid([0, 10], [5, 20], [6, 11])
        l.grid

        [array([0.,1.,2.,3.,4.,5.]),
         array([10.,11.,12.,13.,14.,15.,16.,17.,18.,19.,20.])]
    '''

    def __init__(self, minimum_values, maximum_values, number_of_values, dimnames=None, highervaluesfirst=True):
        '''
        Constructs a linear grid instance based on minimum and maximum 
        values of each variable in each dimension as well as the number 
        of values in each dimension. 
        '''
        # set the instance variables of the linear grid
        self.minval = ensurearray(minimum_values, True)
        self.maxval = ensurearray(maximum_values, True)
        self.nvals = ensurearray(number_of_values, False)
        if (self.nvals <= 0).any(): 
             raise Exception('The number of values building this linear grid must be greater than 0.')
        if dimnames != None: 
            self.dimnames = ensurelist(dimnames) 
        
        # ensure all are the same number of dimensions
        self.ndims = max(self.minval.size, self.maxval.size, self.nvals.size)
        self.dims = range(self.ndims)
        if self.minval.size < self.ndims: 
            if self.minval.size != 1: 
                raise Exception("The maximum, minimum, and number of values arrays as well as the dimnames list must be of same size or 1.")
            self.minval = np.repeat(self.minval, self.ndims)
        if self.maxval.size < self.ndims: 
            if self.maxval.size != 1: 
                raise Exception("The maximum, minimum, and number of values arrays as well as the dimnames list must be of same size or 1.")
            self.maxval = np.repeat(self.maxval, self.ndims)
        if self.nvals.size < self.ndims: 
            if self.nvals.size != 1: 
                raise Exception("The maximum, minimum, and number of values arrays as well as the dimnames list must be of same size or 1.")
            self.nvals = np.repeat(self.nvals, self.ndims)
        if dimnames != None and len(self.dimnames) < self.ndims: 
            if len(self.dimnames) != 1: 
                raise Exception("The maximum, minimum, and number of values arrays as well as the dimnames list must be of same size or 1.")
            self.dimnames = [self.dimnames[0] + '_' + str(i + 1) for i in self.dims]
            
        # calculate the delta value and set the grid list
        self.delta = np.ones(self.ndims)
        self.delta[self.nvals == 1] = 1 
        ne1 = (self.nvals != 1)
        self.delta[ne1] = (self.maxval[ne1] - self.minval[ne1]) / (self.nvals[ne1] - 1)
        
        self.highfirst = highervaluesfirst
        if self.highfirst: 
            # higher values first ensures that states with a high value will be selected when there are value function ties 
            self.grid = [np.linspace(self.maxval[i], self.minval[i], self.nvals[i]) for i in self.dims]
        else: 
            # lower values first ensures that states with a low value will be selected when there are value function ties 
            self.grid = [np.linspace(self.minval[i], self.maxval[i], self.nvals[i]) for i in self.dims]
    
        # get all the possible combinations across dimensions
        self.combos = combinatorics.allcombos(self.grid)
        self.ncombos = len(self.combos)
        self.comboindices = np.arange(self.ncombos)
        self.combovalues = np.zeros((self.ncombos))
        
        # stack lengths up to the beginning in order to index the combos 
        # arrays quickly 
        self.dimlen = np.int32([self.nvals[i:].prod() for i in self.dims if i > 0] + [1])
    
    def argsnap(self, newvalues):
        '''
        Gets the indices within the linear grid that are closest to 
        values within the newvalues array.
        
        See snap() for instructions on how to define newvalue
        
        This method returns the indices and the locations of the indices 
        that are out-of-bounds of this linear grid instance  
        ''' 
        if not(hasattr(newvalues, '__len__') or len(newvalues) != self.ndims): 
            raise Exception("newvalues must be a list with the same number of elements (these elements can be arrays) as the number of dimensions in the linear grid (%d)." % self.ndims)
        
        # get the indices of all the new values within the new grid
        if self.highfirst: 
            i = np.int32(list(np.round((self.maxval - newvalues) / self.delta)))
        else: 
            i = np.int32(list(np.round((newvalues - self.minval) / self.delta)))
        
        # if any are out of bounds, get the indices of those out of bounds ones 
        belowvals = (i < 0)
        abovevals = (i > (self.nvals - 1))
        nanvals = (belowvals + abovevals)
        i[nanvals] = 0
        return (i, nanvals) 
    
    def checksizes(self, indices, nanvalues=None):
        # check the number of columns against dimensions of this linear grid 
        if (len(indices.shape) >= 2 and indices.shape[1] != self.ndims) or (nanvalues != None and len(nanvalues.shape) >= 2 and nanvalues.shape[1] != self.ndims): 
            raise Exception("The indices and nanvalues arrays must have the same number of columns as there are dimensions in this linear grid (%d)" % self.ndims)        
        
    def getgridvalues(self, indices, nanvalues=None):
        '''
        Retrieves the grid values at specified indices. The indices 
        array should be a 2D array  
        '''
        self.checksizes(indices, nanvalues)
        
        # ensure that the indices are integers
        indices = np.int32(list(indices)) 

        # get the snapped values via the indices
        if indices.ndim == 1: 
            retvals = np.float64([self.grid[ind][indices[ind]] for ind in self.dims]).T
        else: 
            retvals = np.float64([self.grid[ind][indices[:, ind]] for ind in self.dims]).T
        
        # set all those out of bounds values with NaN
        if nanvalues != None and nanvalues.any(): 
            retvals[nanvalues] = np.nan

        return retvals
    
    def getflatindices(self, indices):
        # ensure that the indices are integers
        indices = np.int32(indices) 

        # convert the indices of each grid dimension to the index for 
        # the combovalues array        
        # return (indices * self.dimlen).sum(axis=1) # not as fast as np.dot
        return np.dot(indices, self.dimlen)
        
    def getcombovalues(self, indices, nanvalues=None, return_indices=False):
        '''
        Gets the values associated with the combos at the locations of 
        indices... the indices array should have the same number of 
        columns as the number of dimensions (LinGrid.ndims) in the 
        linear grid
        '''
        self.checksizes(indices, nanvalues)

        # get flattened indices that can index combos and combovalues 
        indices = self.getflatindices(indices) 
               
        # get the combo values from the indices using the dimlen array
        combovals = self.combovalues[indices]
        
        # ensure that NaN values are properly represented in the values
        # array  
        if nanvalues != None:
            nanvalues = nanvalues.any(axis=1) 
            combovals[nanvalues] = np.nan
        
        # return indices if desired 
        if return_indices:
            return (combovals, indices, nanvalues)
        else: 
            return combovals
        
    def snap(self, newvalues, return_indices=False):
        '''
        Snaps an array of values to values within the linear grid 
        
        If newvalues is one value set that should be snapped to this 
        linear grid, it must be a list or array of length LinGrid.ndims
        This method will return a 2D numpy array with 1 row and 
        LinGrid.ndims columns 
        
        If newvalues is an array of instances, it must be a 2D numpy 
        array with LinGrid.ndims columns... For example, if 
        LinGrid.ndims is two, newvalues should have two columns. This 
        method will return a 2D numpy array with the same number of rows 
        as newvalues and LinGrid.ndims columns 
        
        A np.nan number is returned when associated with any value in 
        newvalues that is outside of the bounds of the linear grid
        
        If return_indices == True, this method will return a triple 
        that includes (nextstatevalues, indices, nanvalues) where the 
        indices are locations in the state grid at which the next state 
        is evaluated, and nanvalues are the index locations where  
        values were out of bounds of the feasible states     
        ''' 
        # get the indices inside and outside the grid (make sure that 
        # those indices outside of the grid are made NaN in the final 
        # returned values) 
        (i, nanvals) = self.argsnap(newvalues)
         
        # get the grid values at the specified indices  
        retvals = self.getgridvalues(i, nanvals)
        
        # return the grid values as well as the indices if desired
        if return_indices: 
            return (retvals, i, nanvals)
        else: 
            return retvals 

def allcombos(states, decisions):
    '''
    Get all the combinations of possible states and decisions in a big 
    list
    ''' 
    return [(state, decision) for decision in decisions for state in states]

def get_next_states(statefxn, nextstategrid, decisiongrid,
                    currstate, currt, rndvar_realization): 
    '''
    Get an array of all the next states determined from a list of 
    decisions at the current state (performs forward pass). If statefxn 
    calculates previous state based on a previous decision from the 
    current state, then this could potentially be used to perform 
    backward pass.  
    '''
    # get all of the next states in a 2D array
    nextstates = np.array([np.array(statefxn(currstate, dec, currt, rndvar_realization)) for dec in decisiongrid.combos]).reshape((-1, nextstategrid.ndims)) 

    # snap all the next states to possible values 
    (nextstates, i, nanvalues) = nextstategrid.snap(nextstates, return_indices=True)
    
    return (nextstates, i, nanvalues) 

def gen_next_states(simfxn, nextstategrid, decisiongrid, opt_decision,
                    currstate, currt, rndvar_realization, rndvar_i, snapvals=True): 
    '''
    Get an array of all the next states determined from a list of 
    random variable realizations and an optimal decision. 
    '''
    # get all of the next states and decisions
    s_and_d = [simfxn(currstate[i], opt_decision[i], currt, rndvar_realization[i]) for i in rndvar_i]
    nextstates = []
    alldecisions = []
    for (states, decisions) in s_and_d:
        nextstates.append(np.array(states).reshape(-1,)) 
        alldecisions.append(np.array(decisions).reshape(-1,))
    nextstates = np.array(nextstates).reshape((-1, nextstategrid.ndims))
    alldecisions = np.array(alldecisions).reshape((-1, decisiongrid.ndims))

    # snap all the next states to possible values 
    if snapvals:
        (nextstates, i, nanvalues) = nextstategrid.snap(nextstates, return_indices=True)
        (alldecisions, dec_i, dec_nans) = decisiongrid.snap(alldecisions, return_indices=True)
    else:
        (_, i, nanvalues) = nextstategrid.snap(nextstates, return_indices=True)
        (_, dec_i, dec_nans) = decisiongrid.snap(alldecisions, return_indices=True)
        
    return (nextstates, i, nanvalues, alldecisions, dec_i, dec_nans)

def get_values_at_next_state(statefxn, nextstategrid,
                             decisiongrid, currstate, currt,
                             rndvar_realization):
    '''
    Gets the values at the next state given the current state and all 
    possible decisions (in decisiongrid) 
    
    Returns (states, states_i, states_nan, values, values_i, values_nan)
    
    where: 
        states:     values of each state variable for a given decision 
                    (corresponding to a row) 
        states_i:   indices for the state variable within the stategrid
        states_nan: boolean array showing indices at which the state 
                    variable is out-of-bounds
        values:     the value function values at each state 
        values_i:   the indices at which the value function is at a 
                    particular state
        values_nan: boolean array showing indices at which the state 
                    variable is out-of-bounds
    '''
    
    # get all the next states for all decision combinations
    (states, states_i, states_nan) = get_next_states(statefxn,
            nextstategrid, decisiongrid, currstate, currt,
            rndvar_realization)
    
    # get all the values at all the next (discretized) states
    (values, states_i_flat, states_nan_flat) = \
                      nextstategrid.getcombovalues(
                          states_i, states_nan, return_indices=True)
        
    # return these values, indices and nanvalues
    return (states, states_i, states_nan,
            values, states_i_flat, states_nan_flat)

def get_values_at_single_state(valuefxn, statefxn,
                               nextstategrid, decisiongrid,
                               currstate, currt, discount=1,
                               rndvar_space=None, transitionP=None, 
                               acceptablerisk=0.):
    '''
    Gets the values for all decisions at a specific state
    ''' 

    all_next_states_i = [] 
    all_probs = [] 
    rndvars, rndvar_i = get_rndvars(rndvar_space, currstate, currt)
    
    # get the value function at a particular realization of a random variable
    def valuefxn_singlerndvar(rndvar_index):
        # get the random variable realizations 
        rvars = None if (rndvars == None) else rndvars[rndvar_index]

        # get the values at the next state 
        (_, _, _, values_nextstate, next_states_i, _) = \
            get_values_at_next_state(statefxn, nextstategrid,
                                     decisiongrid, currstate, currt,
                                     rvars)
        next_states_i = np.float64(next_states_i) 
         
        # define a function to get the value at a single decision
        def valuefxn_singledec(decision_index):
            return valuefxn(currstate, 
                            nextstategrid.combos[next_states_i[decision_index]],
                            decisiongrid.combos[decision_index],
                            currt, rvars)
        
        # get the values at the current state and decisions
        get_values_currstate = np.vectorize(valuefxn_singledec)
        values_currstate = get_values_currstate(decisiongrid.comboindices)
        
        # get the total values for the current state 
        total_values = values_currstate + discount * values_nextstate
        if rndvar_space != None: 
            def transitionP_singledec(decision_index):
                return float(transitionP(
                        currstate,
                        nextstategrid.combos[next_states_i[decision_index]],
                        decisiongrid.combos[decision_index],
                        currt, rndvar_index))
            getTranPs = np.vectorize(transitionP_singledec)
            tranPs = getTranPs(decisiongrid.comboindices)
            total_values *= tranPs
            next_states_i *= tranPs
            all_probs.append(tranPs)
            
        # add the next state index to this realization
        all_next_states_i.append(next_states_i)
              
        return total_values 
    
    # produce a 2D array where each row is the values for a particular realization
    if rndvar_space == None: 
        total_values = valuefxn_singlerndvar(0)
        all_next_states_i = np.int32(np.round(all_next_states_i[0]))
    else: 
        total_values = np.array([valuefxn_singlerndvar(i) for i in rndvar_i])

        if acceptablerisk == 0.:
            # sum all columns up... if nans or infs are found, nans or infs are returned 
            total_values = np.sum(total_values, axis=0)

            # warn user...
            if not np.isfinite(total_values).any():
                print 'Warning... no acceptable risk level achieved at state %s.' % str(currstate)  
        else:
            # find where nans, infs, and -infs are... true if infeasible, false if feasible 
            # sum the probability of being infeasible
            all_probs = np.array(all_probs) 
            all_probs[np.isfinite(total_values)] = 0
            prob_of_infeasibility = np.sum(all_probs, axis=0)
            
            # next, calculate values with any level of risk (only returns nan if all values in column are nan) 
            total_values = np.nansum(total_values, axis=0)
            
            # finally, filter out values that don't have acceptable risk by assigning np.nan to them
            total_values[prob_of_infeasibility > acceptablerisk] = np.nan

            # warn user...
            if not np.isfinite(total_values).any():
                print 'Warning... no acceptable risk level achieved at state ' + str(currstate) + '. Minimum probability of infeasibility is %g' % prob_of_infeasibility.min() 

        # get the expected value of the state index... This might not work as expected... 
        all_next_states_i = np.int32(np.round(np.array(all_next_states_i).sum(axis=0).round()))
    
    return (all_next_states_i, total_values)

def get_best_value_at_single_state(valuefxn, statefxn,
                                   nextstategrid, decisiongrid,
                                   currstate, currt, discount=1, 
                                   rndvar_space=None, transitionP=None,
                                   ismax=True, verbose=True, acceptablerisk=0.): 
    '''
    Get the best decision and its next state and value from a specific 
    state in the current stage
    
    Returns the best decision index, the index of the next state, 
    and the value for that decision
    ''' 
    global numofprints, maxnumofprints, \
           numoflargeevals, maxnumoflargeevals, evaltimes
    
    starttime = time.time()
    (next_states_i, values) = get_values_at_single_state(
                                           valuefxn, statefxn,
                                           nextstategrid, decisiongrid,
                                           currstate, currt, discount, 
                                           rndvar_space, transitionP,acceptablerisk)
    endtime = time.time() 

    # estimate evaluation time... 
    if numoflargeevals <= maxnumoflargeevals and \
        endtime - starttime > maxsecondsperstage / nextstategrid.ncombos:
        
        # calculate evaltime 
        if evaltimes == None:   
            evaltimes = []
        evaltime = endtime-starttime
        evaltimes.append(evaltime)
        
        # tell user that there's a computation time problem 
        numoflargeevals += 1 
        if numoflargeevals > maxnumoflargeevals and evaltimes != None:
            evaltimes = np.array(evaltimes)
            meaneval = evaltimes.mean()
            stagetime = meaneval * nextstategrid.ncombos / 60.
            print '\n'
            print '********************************************************' 
            print 'Evaluation time to enumerate all decisions over ' \
                + '%d states\n' % len(evaltimes) \
                + 'took an average of %.2f seconds per state.\n' % meaneval
            print 'This indicates that it may take about\n' \
                + '%.1f minutes to complete each stage.' % (stagetime) 
            print '********************************************************\n'
             
    # max or min? 
    if ~np.isnan(values).all(): 
        if ismax: 
            decision_index = np.nanargmax(values)
        else: 
            decision_index = np.nanargmin(values)
        best_value = values[decision_index]
        next_state_i = next_states_i[decision_index]
    else:  
        decision_index = np.nan
        best_value = np.nan
        next_state_i = np.nan
    
    # check for duplicates of the best decision
    if verbose and numofprints <= maxnumofprints:
        tiedbest = (best_value == values) 
        if tiedbest.sum() > 1:
            print "In looping through decisions at time %g, the best value %g was tied %d times:" % (currt, best_value, tiedbest.sum())
            numofprints += 1
            if numofprints > maxnumofprints: 
                print "Number of ties exceeded %d, stopped printing them." % maxnumofprints
    
    # return the best value and its index (for tracking the best answers) 
    return (decision_index, next_state_i, best_value)

def get_best_values_at_current_stage(valuefxn, statefxn,
                                     stategrid, nextstategrid,
                                     decisiongrid,
                                     currt, discount=1, rndvar_space=None,
                                     transitionP=None, ismax=True,
                                     verbose=True,acceptablerisk=0.,
                                     nthreads=1):
    '''
    Determines the values for all the states in the current stage using 
    backward recursion in the non-inverse form 

    Input: 
        valuefxn:      The function that calculates the value for a 
                       specific state, next state, decision, time, and random
                       variable realization
        statefxn:      The function that calculates the next state given 
                       a specified state, decision, and time  
        stategrid:     A linear grid of feasible states
        decisiongrid:  A linear grid of feasible decisions
        currt:         The current timestep
        discount:      A future discount value
        transitionP:   The transition probability function for a 
                       specific current state, next state, decision, 
                       and time
        ismax:         A boolean specifying whether the objective  
                       problem is to maximize (True) or minimize (False)
        verbose:       Specifies whether the methods should be verbose 
                       or not 
    
    Returns (bestdecisions, nextstates, values):  
        bestdecisions:  The index for the best decision at each state in
                        the current stage (currt)
        nextstates:     The index for the next state given the decision
                        taken at the current state 
        values:         The best value at each state in the current 
                        stage (currt) 
    '''
    if nthreads > 1:
        raise Exception("Cannot run multiple threads, they do not work yet.")
    if True: # nthreads <= 1: --> threads are not working for some reason... 
        currvals = np.array([get_best_value_at_single_state(
                                    valuefxn, statefxn,
                                    nextstategrid, decisiongrid,
                                    stategrid.combos[state_index], currt,
                                    discount, rndvar_space, transitionP,
                                    ismax, verbose,
                                    acceptablerisk) for state_index in 
                                    stategrid.comboindices])
    else: 
        # build the initial currvals array 
        currvals = np.zeros((stategrid.ncombos, 3))
        
        # define a function to add only one state at a time to the currvals 
        #  array 
        def run_single_state(state_index):
            currvals[state_index, :] = get_best_value_at_single_state(
                                    valuefxn, statefxn,
                                    nextstategrid, decisiongrid,
                                    stategrid.combos[state_index], currt,
                                    discount, rndvar_space, transitionP,
                                    ismax, verbose,
                                    acceptablerisk)
            
        # start each task in a thread pool and wait for their completion 
        pool = ThreadPool(nthreads)
        for state_index in stategrid.comboindices: 
            pool.add_task(run_single_state, state_index)
        pool.wait_completion()
            
    # get decisions, states, and values from the output
    vals = currvals[:, 2]
    
    # check that there is at least one feasible value for the current set of decisions
    if not np.isfinite(vals).any(): 
        raise Exception('No feasible values found! Consider incorporating some higher acceptable risk level.')
    
    # decision indices (currvals[:,0])
    # next state indices (currvals[, 1]) 
    # best values (currvals[:,2])
    dec_i = np.int32(currvals[:, 0])
    sta_i = np.int32(currvals[:, 1])
    return (dec_i, sta_i, vals) 

def get_grid(grid, currt): 
    '''
    Get the grid that is feasible for a specified time period
    '''
    if isinstance(grid, LinGrid): 
        return grid
    elif hasattr(grid, '__call__'): 
        return grid(currt)
    else: 
        raise Exception('The type of grid given is not a LinGrid or a function of time. Please specify the grid correctly.') 

def get_rndvars(rndvar_space, currstate, currt):
    # get the most current random variable space array
    rndvars = None  
    if rndvar_space == None: 
        return None, None
    if hasattr(rndvar_space, '__call__'): 
        rndvars = rndvar_space(currstate, currt)
    elif isinstance(rndvar_space, np.ndarray): 
        rndvars = rndvar_space
    if rndvars == None or not isinstance(rndvars, np.ndarray) or rndvars.ndim != 2: 
        raise Exception('rndvar_space can only be defined as either a 2D numpy array or a function that returns a 2D numpy array')
    rndvar_i = range(len(rndvars))
    return rndvars, rndvar_i 
    
def generate_rndvar(rndvar_space, transitionP, size, 
                    currstate, nextstate, currdecision, currt):
    # generate a random variable
    rndvars, rndvar_i = get_rndvars(rndvar_space, currstate, currt)
    def gen_one_random(sceni):
        probabilities = np.array([float(transitionP(currstate[sceni], nextstate[sceni], currdecision[sceni], currt, i)) for i in rndvar_i])
        bins = np.add.accumulate(probabilities)
        return rndvars[np.digitize(random_sample(1), bins)[0]]
    return np.array([gen_one_random(i) for i in range(size)])

def check_traceinputs(rndvar_space, rndvar_realizations, transitionP, simfxn, valuefxn):
    check_inputs(rndvar_space, transitionP)
    if rndvar_space != None and (transitionP == None or simfxn == None or valuefxn == None): 
        raise Exception('If the random variable matrix is defined, the transitionP and simfxn variables must also be defined.')
    if rndvar_space != None and rndvar_realizations != None: 
        raise Exception('If the rndvar_space is defined, rndvar_realizations cannot also be defined') 
    if simfxn != None and (rndvar_space == None and rndvar_realizations == None): 
        raise Exception('If the simfxn variable is defined, either rndvar_space or rndvar_realizations must also be defined.')
    
def trace_optimal_paths(stategrid, decisiongrid, timelist, state0,
                    optimal_results,
                    ismax=True, verbose=True, discount=1, 
                    rndvar_space=None, rndvar_realizations=None, 
                    transitionP=None, valuefxn=None, simfxn=None, 
                    nscenarios=200, errorOnInfeasible=True, snapvals=True):
    '''
    Simulate the optimized system in a forward pass motion

    Parameters:
    @param stategrid:      
                    A LinGrid instance of discretized possible states which are 
                    internally held in a 2D array with the following structure: 
                    
                    example: 
                        np.array([[1,2,3],
                                  [1,2,5],
                                  [1,4,3],
                                  [1,4,5],
                                  [3,2,3],
                                  [3,2,5],
                                  [3,4,3],
                                  [3,4,5]])
    
                    stategrid may also be a function that returns a LinGrid 
                    instance given the current timestep as the input argument   
    
    @param decisiongrid:   
                    A LinGrid instance of discretized possible states which are 
                    internally held in a 2D array with the following structure:
                    
                    example: 
                        see example in states above
    
                    decisiongrid may also be a function that returns a LinGrid 
                    instance given the current timestep as the input argument   
    
    @param timelist:       
                    A list of (sorted) discrete times. 
    
    @param state0:         
                    The initial state vector (size n)
    
    @param optimal_results: 
                    A dictionary containing the optimal results from 
                    the solve_* functions. The dictionary usually contains 
                    the following keys specifying the type of data: 
                    
                    opt_decisions_indices:      
                        Array containing indices for optimal decisions at each 
                        current state
                    opt_next_states_indices: 
                        Array containing indices for optimal next states at each 
                        current state
                    opt_decisions:
                        Array containing optimal decision vectors for each 
                        current state
                    opt_next_states: 
                        Array containing optimal state vectors for each current 
                        state
                    opt_values: 
                        Array containing optimal values for each current state 

    @param ismax:          
                    Specifies whether the optimization problem is a 
                    maximization problem or a minimization problem 
                    
                    Possible values: True or False

    @param verbose:        
                    Specifies whether the optimization routine should be verbose
                    for debugging purposes 
                    
                    Possible values: True or False

    @param discount (scalar):       
                    A discount factor for future values. If not 
                    specified, discount = 1 (no discount of future 
                    periods)
    
    @param rndvar_space:
                    A 2D array of all possible random variable(s) realizations
                    one set of realizations per row, or a function that returns 
                    such an array with the following usage: 
                    
                    rndvar_space(curr_state, beg_period_time)
     
    @param rndvar_realizations: 
                    If wanting to simulate a known trace of random variable 
                    realizations, this is an array of random variable 
                    realizations with the same number of elements as time 
                    periods minus 1 (the last time is the instantaneous 
                    beg-of-period time) 

    @param transitionP:    
                    The transition probability function for a 
                    specific current state, next state, decision, 
                    beg-of-period time t, and a realization of the random 
                    variable (if specified)
                    
                    transitionP(curr_state, next_state, curr_decision,
                                end_period_time, rndvar_realization)

    @param valuefxn:       
                    A function that calculates the value function  
                    based on the current state vector (size n), 
                    decision vector (size m), beg-of-period time t, and a 
                    realization of a random variable
                    
                    valuefxn(curr_state, next_state, curr_decision,
                             beg_period_time, rndvar_realization)
    
    @param simfxn:         
                    A function that performs the same thing that the statefxn 
                    does, except that if there is any infeasibility, it corrects 
                    the decision in order to calculate the next state still... 
                    Thus, this function returns a tuple (next_state, decision)  
                    
    @param nscenarios:     
                    The number of scenarios to perform in a stochastic run
                    default is 200
                        
    @param errorOnInfeasible: 
                    A boolean flag specifying whether or not to raise an error 
                    when an infeasible state has occurred. Default is True. 
                    
    @param snapvals:
                    A boolean flag specifying whether or not to snap values 
                    when calculating the next state during while performing 
                    forward trace 

    @return output dictionary:
                    A dictionary with the following keys specifying the various 
                    aspects of the traced optimization results: 
                    
                    trace_decisions
                        Array containing the traced optimal decision path 
                    trace_states 
                        Array containing the traced state path from state0
                    trace_values 
                        Array containing final traced value path 
                    trace_incrvalues
                        Array containing final incremental traced value path
                    trace_rndvars 
                        When random variable realizations are present, this key 
                        will appear in the output dictionary and contains the 
                        realizations of all the random variables while the 
                        tracing occurred
    '''
    # run checks 
    check_traceinputs(rndvar_space, rndvar_realizations, transitionP, simfxn, valuefxn)

    # optimal results from the optimization
    opt_decision_indices = optimal_results['opt_decisions_indices']
    opt_nextstates_indices = optimal_results['opt_next_states_indices'] 
    opt_values = optimal_results['opt_values']
    incr_values = None 
    
    # if the optimal values are in an array (as opposed to a list), they 
    # represent the infinite horizon policy, not the optimal decisions at a 
    # particular time 
    ispolicy = (isinstance(opt_values, np.ndarray) and len(opt_values.shape) == 1) 
                
    # check the initial state and get indices 
    currt = timelist[0]
    sgrid = get_grid(stategrid, currt)
    state0, state0_i, state0_nan = sgrid.snap(ensurelist(state0), return_indices=True)
    state0 = state0.reshape((-1,))
    state0_i = sgrid.getflatindices(state0_i).reshape((-1,))
    if state0_nan.any():
        raise Exception("One of the initial states is not feasible.")

    # get random variable indices 
    if rndvar_space != None: 
        if len(state0_i) != 1:
            raise Exception("When generating random simulations, must specify one and only one initial state.")
        state0 = np.tile(state0, (nscenarios,1))
        state0_i = np.tile(state0_i, (nscenarios,1))

    # prepare output lists 
    traceopt_decisions = [] 
    traceopt_states = []
    traceopt_values = []
    traceopt_rndvars = [] 
    traceopt_incrvalues = []
    traceopt_states.append(state0)
    if ispolicy: 
        traceopt_values.append(opt_values[state0_i])
    else: 
        traceopt_values.append(opt_values[0][state0_i])
    
    # set up the times array 
    def getoptvalues(state_i, state, currt_index):
        # initialize parameters
        rndvars = None
        
        # get the state and decision grids 
        currt = timelist[currt_index]
        sgrid = get_grid(stategrid, currt)
        dgrid = get_grid(decisiongrid, currt)
        
        # get optimal decision
        if ispolicy:
            opt_decision_i = opt_decision_indices[state_i] if sgrid.ncombos != 1 else opt_decision_indices.repeat(state_i.size)
            opt_decision_i.resize((opt_decision_i.size))
        else: 
            opt_decision_i = opt_decision_indices[currt_index][state_i] if sgrid.ncombos != 1 else opt_decision_indices[currt_index].repeat(state_i.size)
            opt_decision_i.resize((opt_decision_i.size))
        opt_decision = dgrid.combos[opt_decision_i]
        if dgrid.ndims == 1: 
            opt_decision = opt_decision.reshape((-1))

        # get optimal (or expected) next state 
        if ispolicy: 
            opt_nextstate_i = opt_nextstates_indices[state_i] if sgrid.ncombos != 1 else opt_nextstates_indices.repeat(state_i.size)
        else:   
            opt_nextstate_i = opt_nextstates_indices[currt_index][state_i] if sgrid.ncombos != 1 else opt_nextstates_indices[currt_index].repeat(state_i.size)
        opt_nextstate = sgrid.combos[opt_nextstate_i]
        if opt_nextstate.ndim > 1: 
            opt_nextstate = opt_nextstate.reshape((-1,))
        
        # generate next state during simulation            
        if simfxn != None: 
            if rndvar_space != None:
                # generate random variables for this timestep
                rndvars = generate_rndvar(rndvar_space, transitionP, 
                                          nscenarios, state, opt_nextstate, 
                                          opt_decision, currt)
            elif rndvar_realizations != None:
                # use the provided set of random variables 
                rndvars = rndvar_realizations[currt_index]

            # get all next states and decisions considering the randomly 
            # generated values
            opt_nextstate, opt_nextstate_i, s_nans, \
            opt_decision, opt_decision_i, d_nans = gen_next_states(
                                          simfxn, sgrid, dgrid, 
                                          opt_decision, state, currt, 
                                          rndvars, range(nscenarios),
                                          snapvals=snapvals)
#             if opt_nextstate.ndim > 1: 
#                 opt_nextstate = opt_nextstate.reshape((-1))
            
            # errors if any nan values found
            if s_nans.any() or (errorOnInfeasible and d_nans.any()): 
                raise Exception("When simulating the system, cannot generate infeasible states or decisions...\n\nState NANs:\n%s\n\nDecision NANs:\n%s\n\n" % (s_nans, d_nans))
        
            # calculate objective function value using valuefxn
            if ispolicy: 
                opt_value = opt_values[opt_nextstate_i]
            else:   
                opt_value = opt_values[currt_index + 1][opt_nextstate_i]
            incr_value = np.array([valuefxn(state[i],opt_nextstate[i],opt_decision[i],currt,rndvars[i]) for i in range(nscenarios)]).reshape((-1,1))
        
        else: 
                
            # get optimal values at the next state
            if ispolicy: 
                opt_value = opt_values[opt_nextstate_i]
                incr_value = opt_values[state_i] - discount * opt_value
            else:   
                opt_value = opt_values[currt_index + 1][opt_nextstate_i]
                incr_value = opt_values[currt_index][state_i] - discount * opt_value
        
        return (opt_decision, opt_nextstate, opt_value, opt_nextstate_i, rndvars, incr_value)
    
    # trace the optimal path through the entire decision period 
    state_i = state0_i
    state = state0
    time_i = range(len(timelist)-1) 
    for i in time_i:  # skip the first state (already determined)
        
        # get optimal values within the current time step
        # (these include the optimal decisions selected to get us here, 
        #  the states arrived to given optimal decisions, and the 
        #  objective function values at those locations)
        (opt_decision, opt_nextstate, opt_value, 
         opt_nextstate_i, rndvars, incr_value) = getoptvalues(state_i, state, i)
            
        # update the output lists
        traceopt_decisions.append(opt_decision)        
        traceopt_states.append(opt_nextstate)
        traceopt_values.append(opt_value)
        traceopt_incrvalues.append(incr_value)
        if rndvars != None: 
            traceopt_rndvars.append(rndvars)
        
        # update the previous state index
        state_i = opt_nextstate_i
        state = opt_nextstate

    # calculate the present values
    if simfxn != None: 
        time_i.reverse()
        for i in time_i: 
            traceopt_values[i][:] = traceopt_incrvalues[i] + discount * traceopt_values[i + 1]
            
    # print out results if wanted 
    if verbose: 
        print 'Optimal values:'
        print traceopt_values
        print 'Optimal decisions:'
        print traceopt_decisions
        print 'Optimal states:'
        print traceopt_states
        print 'Random variable realizations:'
        print traceopt_rndvars
    
    # return decisions for each timestep
    output = {} 
    output['trace_decisions'] = traceopt_decisions
    output['trace_states'] = traceopt_states
    output['trace_values'] = traceopt_values
    output['trace_incrvalues'] = traceopt_incrvalues
    if rndvars != None:  
        output['trace_rndvars'] = traceopt_rndvars
    return output

def check_inputs(rndvar_space, transitionP): 
    if rndvar_space != None: 
        if isinstance(rndvar_space, np.ndarray) and rndvar_space.ndim != 2: 
            raise Exception("The random variable array must have two dimensions, where each column represents a different random variable, and each row a different realization.")
        if transitionP == None: 
            raise Exception("If the random variable array is defined, the transition probabilities (transitionP) must be defined.")
    else: 
        if transitionP != None: 
            raise Exception("If transition probabilities are defined, the random variable array must also be defined.")
        
def solve_finitetime(valuefxn, statefxn, stategrid, decisiongrid,
                     timelist,
                     discount=1, rndvar_space=None, transitionP=None,
                     terminalValues=0, ismax=True, verbose=True,
                     acceptablerisk=0.,nthreads = 1):        
    '''
    Solves a finite time dynamic programming problem where the number 
    of states is n and decisions is m with the following parameters:  
    
    Parameters:
    @param valuefxn:       
                    A function that calculates the value function  
                    based on the current state vector (size n), next state 
                    vector (size n), decision vector (size m), 
                    beginning-of-period time t, and a realization of a random 
                    variable
                    
                    valuefxn(curr_state, next_state, curr_decision,
                             beg_period_time, rndvar_realization)
    
    @param statefxn (size n):       
                    A function that calculates the next state given  
                    the current state vector (size n), decision vector 
                    (size m), beg-of-period time t, and a realization of a 
                    random variable 
               
                    statefxn(curr_state,curr_decision,beg_period_time,
                             rndvar_realization)
    
    @param stategrid:      
                    A LinGrid instance of discretized possible states which are 
                    internally held in a 2D array with the following structure: 
                    
                    example: 
                        np.array([[1,2,3],
                                  [1,2,5],
                                  [1,4,3],
                                  [1,4,5],
                                  [3,2,3],
                                  [3,2,5],
                                  [3,4,3],
                                  [3,4,5]])
    
                    stategrid may also be a function that returns a LinGrid 
                    instance given the current timestep as the input argument   
    
    @param decisiongrid:   
                    A LinGrid instance of discretized possible states which are 
                    internally held in a 2D array with the following structure:
                    
                    example: 
                        see example in states above
    
                    decisiongrid may also be a function that returns a LinGrid 
                    instance given the current timestep as the input argument   
    
    @param timelist:       
                    A list of (sorted) discrete times. 
    
    @param discount (scalar):
                    A discount factor for future values. If not 
                    specified, discount = 1 (no discount of future 
                    periods)
    
    @param rndvar_space:
                    A 2D array of all possible random variable(s) realizations
                    one set of realizations per row, or a function that returns 
                    such an array with the following usage: 
                    
                    rndvar_space(curr_state, beg_period_time)
     
    @param transitionP:    
                    The transition probability function for a 
                    specific current state, next state, decision, 
                    beg-of-period time t, and a realization of the random 
                    variable (if specified)
                    
                    transitionP(curr_state, next_state, curr_decision,
                                end_period_time, rndvar_realization)
                    
    @param terminalValue:  
                    The terminal value vector for each state realization, which 
                    has the same number of elements as the number of 
                    combinations of all discretized state values 
                    (size N_1 * N_2 * ... * N_n) where N_i is the number of 
                    discretized values for the i-th state variable
                    
    @param ismax:          
                    Specifies whether the optimization problem is a 
                    maximization problem or a minimization problem 
                    
                    Possible values: True or False
    
    @param verbose:        
                    Specifies whether the optimization routine should be verbose
                    for debugging purposes 
                    
                    Possible values: True or False
    
    @param acceptablerisk: 
                    Specifies the level at which there is an acceptable level of 
                    infeasibility 
    
    @param nthreads:       
                    Currently an unused variable since multithreading does not 
                    work, nor does it speed up execution (this can only be 1)

    @return output dictionary:
                    A dictionary with the following keys specifying the various 
                    aspects of the optimization results: 
                    
                    opt_decisions_indices:      
                        Array containing indices for optimal decisions at each 
                        current state
                    opt_next_states_indices: 
                        Array containing indices for optimal next states at each 
                        current state
                    opt_decisions:
                        Array containing optimal decision vectors for each 
                        current state
                    opt_next_states: 
                        Array containing optimal state vectors for each current 
                        state
                    opt_values: 
                        Array containing optimal values for each current state 
    '''
    # input check 
    check_inputs(rndvar_space, transitionP)

    # reset global variables 
    global numofprints
    numofprints = 0

    # prepare output information
    nstages = len(timelist) 
    opt_decisions_indices = [] 
    opt_next_states_indices = []
    opt_decisions = [] 
    opt_next_states = []
    opt_values = []
    
    # initialize the terminal values
    sgrid = get_grid(stategrid, timelist[nstages - 1])
    sgrid.combovalues[:] = terminalValues
    opt_values.append(copy(sgrid.combovalues))

    # perform backward pass 
    timelist.reverse()  # go backwards in time   
    nextt = timelist[0]
    for currt in timelist[1:]:  # the last period is the terminal period (already determined)
        sys.stdout.write('\rSolving...  Timestep %d       ' % currt)
        sys.stdout.flush() 
        
        # get the current grids  
        sgrid = get_grid(stategrid, currt)
        dgrid = get_grid(decisiongrid, currt)
        nextsgrid = get_grid(stategrid, nextt)
        
        # get all the best values at the current stage
        decision_i, next_state_i, vals = get_best_values_at_current_stage(
                                         valuefxn, statefxn,
                                         sgrid, nextsgrid, dgrid,
                                         currt, discount,
                                         rndvar_space, transitionP, ismax,
                                         verbose,acceptablerisk,nthreads)
        
        # update the grid values 
        sgrid.combovalues[:] = vals  # must be the same size
        
        # fill output lists 
        opt_decisions_indices.append(decision_i) 
        opt_next_states_indices.append(next_state_i) 
        opt_values.append(vals)
        opt_decisions.append(dgrid.combos[decision_i])
        opt_next_states.append(sgrid.combos[next_state_i])
        
        # update next period
        nextt = currt
    print '\rSolving...  Done!           '
    
    # reverse the lists to get the time in order 
    opt_decisions_indices.reverse()
    opt_next_states_indices.reverse() 
    opt_values.reverse()
    opt_decisions.reverse()
    opt_next_states.reverse()
    timelist.reverse() 
    
    output = {} 
    output['opt_decisions_indices'] = np.array(opt_decisions_indices)
    output['opt_next_states_indices'] = opt_next_states_indices
    output['opt_decisions'] = opt_decisions
    output['opt_next_states'] = opt_next_states
    output['opt_values'] = opt_values
    return output

def solve_infinitetime(valuefxn, statefxn, stategrid, decisiongrid, tolerance,
                     discount=1, rndvar_space=None, transitionP=None,
                     ismax=True, verbose=True, maxiterations=100,
                     acceptablerisk=0.,nthreads=1):
    '''
    Solves an infinite time horizon dynamic programming problem where the number 
    of states is n and decisions is m with the following parameters:  
    
    Parameters:
    @param valuefxn:       
                    A function that calculates the value function  
                    based on the current state vector (size n), next state 
                    vector (size n), decision vector (size m), 
                    beginning-of-period time t, and a realization of a random 
                    variable
                    
                    valuefxn(curr_state, next_state, curr_decision,
                             beg_period_time, rndvar_realization)
    
    @param statefxn (size n):       
                    A function that calculates the next state given  
                    the current state vector (size n), decision vector 
                    (size m), beg-of-period time t, and a realization of a 
                    random variable 
               
                    statefxn(curr_state,curr_decision,beg_period_time,
                             rndvar_realization)
    
    @param stategrid:      
                    A LinGrid instance of discretized possible states which are 
                    internally held in a 2D array with the following structure: 
                    
                    example: 
                        np.array([[1,2,3],
                                  [1,2,5],
                                  [1,4,3],
                                  [1,4,5],
                                  [3,2,3],
                                  [3,2,5],
                                  [3,4,3],
                                  [3,4,5]])
    
    @param decisiongrid:   
                    A LinGrid instance of discretized possible states which are 
                    internally held in a 2D array with the following structure:
                    
                    example: 
                        see example in states above
    
    @param tolerance:      
                    Specifies the tolerance at which the routine should stop 

    @param discount (scalar):       
                    A discount factor for future values. If not 
                    specified, discount = 1 (no discount of future 
                    periods)
    
    @param rndvar_space:   
                    A 2D array of all possible random variable(s) realizations
                    one set of realizations per row, or a function that returns 
                    such an array with the following usage: 
                    
                    rndvar_space(curr_state, beg_period_time)
     
    @param transitionP:    
                    The transition probability function for a 
                    specific current state, next state, decision, 
                    beg-of-period time t, and a realization of the random 
                    variable (if specified)
                    
                    transitionP(curr_state, next_state, curr_decision,
                                end_period_time, rndvar_realization)
                    
    @param ismax:          
                    Specifies whether the optimization problem is a 
                    maximization problem or a minimization problem 
                    
                    Possible values: True or False
    
    @param verbose:        
                    Specifies whether the optimization routine should be verbose
                    for debugging purposes 
                    
                    Possible values: True or False
    
    @param maxiterations:  
                    A number specifying the maximum number of iterations that 
                    the optimization routine will take prior to forcing an exit 
                    of the optimization routine 
                    
                    Possible values: 0 to Large Integer 
     
    @param acceptablerisk: 
                    Specifies the level at which there is an acceptable level of 
                    infeasibility 
    
    @param nthreads:       
                    Currently an unused variable since multithreading does not 
                    work, nor does it speed up execution (this can only be 1)

    @return output:
                    A dictionary with the following keys specifying the various 
                    aspects of the optimization results: 
                    
                    opt_decisions_indices:      
                        Array containing indices for optimal decisions at each 
                        current state
                    opt_next_states_indices: 
                        Array containing indices for optimal next states at each 
                        current state
                    opt_decisions:
                        Array containing optimal decision vectors for each 
                        current state
                    opt_next_states: 
                        Array containing optimal state vectors for each current 
                        state
                    opt_values: 
                        Array containing optimal values for each current state
                    maxerrors:
                        The difference between optimal objective function values 
                        from one iteration to the next for all iterations. This 
                        is mainly for convergence checking
    '''
        
    # initialize the values 
    i = 0
    stategrid.combovalues[:] = 0
    vals = copy(stategrid.combovalues)
    prev_vals = copy(stategrid.combovalues) + np.inf
    maxerror = tolerance + 1  
    maxerrors = []
    while i < 2 or (i < maxiterations and maxerror > tolerance): 
        sys.stdout.write('\rSolving...  Iteration %d       ' % i)
        sys.stdout.flush() 
        
        # get all the best values at the current stage
        _, _, vals = get_best_values_at_current_stage(
                                         valuefxn, statefxn,
                                         stategrid, stategrid, decisiongrid,
                                         None, discount,
                                         rndvar_space, transitionP, ismax,
                                         verbose,acceptablerisk,nthreads)
        
        # update current values and previous values 
        prev_vals = copy(stategrid.combovalues) 
        stategrid.combovalues[:] = vals # must be the same size
        
        # curr max error 
        maxerror = np.abs(vals-prev_vals).max()
        maxerrors.append(maxerror)
        
        # update iteration  
        i += 1
    print '\rSolving...  Done!           '
    
    # inform user if not converged on answer
    if i >= maxiterations and maxerror > tolerance: 
        print 'Did not converge on an answer, the max error is %g while the tolerance is %g' % (maxerror, tolerance) 
        
    # get the final results        
    decision_i, next_state_i, vals = get_best_values_at_current_stage(
                                     valuefxn, statefxn,
                                     stategrid, stategrid, decisiongrid,
                                     None, discount,
                                     rndvar_space, transitionP, ismax,
                                     verbose,acceptablerisk)
    decisions = decisiongrid.combos[decision_i]
    nextstates = stategrid.combos[next_state_i]
    
    # assemble the output dictionary
    output = {} 
    output['opt_decisions_indices'] = decision_i 
    output['opt_next_states_indices'] = next_state_i
    output['opt_decisions'] = decisions
    output['opt_next_states'] = nextstates
    output['opt_values'] = vals
    output['maxerrors'] = maxerrors
    return output 
    
def save_solution(thedir, file_prefix, result_dict):
    '''
    Save the dictionary of results to binary numpy files returned from 
    solve_* or trace_* functions
    ''' 
    if not os.path.exists(thedir):
        os.makedirs(thedir)
    for k in result_dict: 
        np.save(os.path.join(thedir, file_prefix + k + '.npy'), result_dict[k])

def load_solution(thedir, file_prefix):
    '''
    Loads a dynamic programming solution that was saved using save_solution().
    
    @return Returns a dictionary containing the results 
    '''
    
    # get the full path 
    thebase = os.path.join(thedir, file_prefix) 
    theptrn = thebase + '*.npy'
    
    # search for files 
    thefiles = glob.glob(theptrn) 
    if thefiles == []: 
        raise Exception('No solution files with the following pattern were found:\n  %s' % theptrn)
    
    # fill a dictionary with the results 
    results = {} 
    for f in thefiles:
        k = os.path.splitext(f[len(thebase):])[0] 
        results[k] = np.load(f)
        
    # return the dictionary
    return results
        
    
    
    
    