'''
Created on Dec. 10, 2013

Performs optimal control of a reservoir without capacity

@author: Andre Dozier
'''

# imports
import numpy as np
import matplotlib.pyplot as plt
import scipy.integrate as scint 
import scipy.optimize as scopt 

# simulate the system of equations 
def simsys(y0, T, syseqs, Nperiods=100):
    t = np.linspace(0, T, Nperiods)
    simy = scint.odeint(syseqs, y0, t)
    return simy

# T and xT given -> choose u0 -> set sim_xT = xT
def simsysT(u0, T, x0, xT, syseqs, Nperiods=100): 
    simy = simsys(np.array([x0, u0]), T, syseqs, Nperiods=Nperiods)
    return (simy[-1, 0] - xT)
def fwdshoot(u0, T, x0, xT, syseqs, Nperiods=100):
    u0opt = scopt.fsolve(simsysT, u0, (T, x0, xT, syseqs, Nperiods))
    simy = simsys(np.array([x0, u0opt]), T, syseqs, Nperiods=Nperiods)  
    simx = simy[:, 0]
    simu = simy[:, 1]
    return (simx, simu)

# T free, xT given -> choose T -> set H = 0
def fwdshoot_T_Free_sub(T, u0, x0, xT, syseqs, Hamiltonian, Nperiods=100): 
    simx, simu = fwdshoot(u0, T, x0, xT)
    return Hamiltonian(simx[-1], simu[-1], T)
def fwdshoot_T_Free(u0, T0, x0, xT, syseqs, Hamiltonian, Nperiods=100): 
    Topt = scopt.fsolve(fwdshoot_T_Free_sub, T0, 
                        (u0, x0, xT, syseqs, Hamiltonian, Nperiods))
    return tuple(list(fwdshoot(u0, Topt, x0, xT, syseqs, Nperiods=Nperiods))\
                  + [Topt])

# # T given, xT free 
# def simsys_xT_Free(L0, T, x0, _):
#     simy = simsys(np.array([x0, L0]), T)
#     return simy[-1, 1]  # returns simulated LT (want this to be zero)
# def fwdshoot_xT_Free(L0, T, x0, _):
#     L0opt = scopt.fsolve(simsys_xT_Free, L0, (T, x0))[0]
#     simy = simsys(np.array([x0, L0opt]), T) 
#     simx = simy[:, 0]
#     simu = simy[:, 1]
#     return (simx, simu)

# phase plane with null clines 
def plotphasediag(xdot, udot, xname, uname, xlims, ulims, xdot0nc, udot0nc, 
                  x0=None, T=None, xT=None, verbose=False, xss=None, uss=None, 
                  u0guess=1, Nperiods=100, Hamiltonian=None,fignum=None, 
                  outfigfile=None):
    if fignum == None: 
        plt.figure() 
    else: 
        plt.figure(fignum) 
     
    # plot directional arrows
    x_all = np.linspace(xlims[0], xlims[1], 25)
    u_all = np.linspace(ulims[0], ulims[1], 25)
    x, L = np.meshgrid(x_all, u_all)
    xdotmat = xdot(x, L, 0) # perform analysis for t=0 if this diff. eq. depends on t
    udotmat = udot(x, L, 0) # perform analysis for t=0 if this diff. eq. depends on t
    Q = plt.quiver(x, L, xdotmat, udotmat, alpha=0.3)
    plt.quiverkey(Q, 0.84, 0.05, 1, r'$1$', labelpos='E', coordinates='figure', 
                  fontproperties={'weight': 'bold'})
    
    # plot isoclines (a.k.a., nullclines) 
    xnc = xdot0nc(x_all)  # get nullcline where xdot = 0
    ync = udot0nc(x_all)  # get nullcline where udot = 0
    plt.plot(x_all, xnc, 'k:', label=r'$\dot{' + xname + '}_t = 0$', 
             linewidth=2)
    plt.plot(x_all, ync, 'k--', label=r'$\dot{' + uname + '}_t = 0$', 
             linewidth=2)

    # prepare output variables
    allsimx = []
    allsimu = [] 
    allt = [] 
    
    if x0 != None:   
        # check xT 
        if xT == None: 
            raise Exception('Must specify x_T, cannot simulate x_T free yet.')

        # prepare plotting variable grayscale intensity
        alpha = 1.
        alphai = 1.
    
        # set up the time periods to run 
        Tgiven = (T != None)
        if T == None: 
            T = [10] # this is an initial time period for the solution
        if not isinstance(T, list): 
            T = [T]

        # set up the system of equations 
        def syseqs(y, t): 
            # y[0] is x (state), y[1] is u (decision)
            return (xdot(y[0], y[1], t), udot(y[0], y[1], t)) 
    
        # simulate each optimal path
        for t in T: 
            if Tgiven: # T is given
                simx, simu = fwdshoot(u0guess, t, x0, xT, syseqs, Nperiods=Nperiods)
                topt = t
            else: # T is free, must have Hamiltonian 
                if Hamiltonian == None: 
                    raise Exception('Must provide the Hamiltonian as a function of state and decision variables in that order when selecting to have T free.')
                simx, simu, topt = fwdshoot_T_Free(u0guess, t, x0, xT, syseqs, Hamiltonian, Nperiods=Nperiods)
            L0opt = simu[0]
            plt.plot([x0, x0], ulims, 'k-.')
            if xT != None: 
                plt.plot([xT, xT], ulims, 'k-.')
            plt.plot(simx, simu, 'k-', alpha=alpha, label=r'Path from (%.2f,%.2f) for T = %g' % (x0, L0opt, topt))
            alpha = 1 - alphai / (len(T) + 1)
            alphai += 1              
            if verbose: 
                print simx
                print simu
            allsimx.append(simx)
            allsimu.append(simu)
            allt.append(topt)
                
    if xss != None: 
        perc5 = 0.05 * (plt.ylim()[1] - plt.ylim()[0])
        if isinstance(xss, list):
            for i in range(len(xss)): 
                plt.annotate('steady state (%g, %g)' % (xss[i], uss[i]), xy=(xss[i], uss[i]), xycoords='data', xytext=(xss[i], uss[i] - perc5), va='center', ha='center', arrowprops=dict(arrowstyle="->"))
        else: 
            plt.annotate('steady state (%g, %g)' % (xss, uss), xy=(xss, uss), xycoords='data', xytext=(xss, uss - perc5), va='center', ha='center', arrowprops=dict(arrowstyle="->"))
    plt.xlabel('$' + xname + '_t$')
    plt.ylabel('$' + uname + '_t$')
    plt.xlim(xlims)
    plt.ylim(ulims)
    # plt.xticks([])
    # plt.yticks([])
    plt.legend(loc='best', fontsize=9)
    if outfigfile != None: 
        plt.savefig(outfigfile)
    return (allsimx, allsimu, allt)
