'''
Created on Dec. 3, 2013

Contains a test for the dynamic programming package

@author: Andre Dozier
'''

import dypro.dypro as dp 
import numpy as np 
# import matplotlib.pyplot as plt

testname = 'Miranda and Fackler (2002) Ex. 7.6.3'
def run(verbose=False, doplots=False): 
    # test from Miranda and Fackler (2002). Applied Computational Economics and Finance. p. 176
    
    # set up model parameters 
    maxage = 5
    repcost = 75.
    mancost = 10.
    delta = 0.9
    n = maxage
    state0 = np.array([1,0])
     
    # set up the state and decision grids and time   
    stategrid = dp.LinGrid([1, 0], [maxage, maxage-1], n)
    decisiongrid = dp.LinGrid(0, 2, 3) # discrete variable (keep = 0, service = 1, replace = 2)
    
    # time 
    t = range(12)

    # maximize or minimize? 
    ismax = True

    # define the state equation
    def statefxn(s, x, t, w):
        if x == 0: # keep
            return np.array([min(s[0] + 1, n), s[1]])
        elif x == 1: # service 
            return np.array([min(s[0] + 1, n), s[1] + 1])
        else: # replace
            return np.array([1, 0])
    
    # define the value function 
    def valuefxn(s, snext, x, t, w):
        if x == 2: # replace 
            return (50 - repcost)
        
        # keep or service
        q = (50 - 2.5 * s[0] - 2.5 * (s[0] ** 2))
        if x == 0: # keep
            return q * min(1., 1-(s[0] - s[1])/float(maxage))  
        else: # service
            return q * min(1., 1-(s[0] - s[1] - 1)/float(maxage)) - mancost
    
    # solve the problem  
    opt_results = dp.solve_finitetime(valuefxn, statefxn, stategrid, 
                decisiongrid, t, ismax=ismax, discount=delta, verbose=verbose)
    
    if verbose: 
        print 'opt_decisions_indices'
        print opt_results['opt_decisions_indices']
        
    # trace optimal paths
    trace_results = dp.trace_optimal_paths(stategrid, decisiongrid, t, state0,
                    opt_results, ismax=ismax, verbose=verbose)            
                    
    # extract desired information             
    decisions = trace_results['trace_decisions']
    states = trace_results['trace_states']
    values = trace_results['trace_values']

    # if doplots: 
        # plt.ion()
        # plt.close('all')
        
        # # plot decisions against time 
        # plt.figure()
        # timearray = np.array(t[:-1]).reshape((-1,1))
        # plt.plot(timearray, decisions, 'k-')
        # plt.xlabel('Time')
        # plt.ylabel('Optimal decision path')
        
        # # plot stock against time 
        # timearray = np.array(t).reshape((-1,1))
        # plt.figure()
        # plt.plot(timearray, states[:,0], 'k-', label='Asset Age')
        # plt.plot(timearray, states[:,1], 'k--', label='Number of Servicings')
        # plt.xlabel('Time')
        # plt.ylabel('Optimal state path')
        # plt.legend(loc='best')
        
        # # plot stock against time 
        # plt.figure()
        # plt.plot(timearray, values, 'k-')
        # plt.xlabel('Time')
        # plt.ylabel('Optimal value path')

        # plt.show() 

    # check optimal decisions
    correct_decisions = np.array([1,1,0,2,1,1,0,2,1,1,0]).reshape((-1, 1))
    correct = (np.abs(decisions - correct_decisions) < 0.0001)
    if not correct.all(): 
        return False
         
    # check optimal values
    s0 = np.array([1,2,3,4,1,2,3,4,1,2,3,4]).reshape((-1, 1))
    s1 = np.array([0,1,2,2,0,1,2,2,0,1,2,2]).reshape((-1, 1))
    states = np.array(states)
    correct_states = np.hstack((s0, s1))
    correct = (np.abs(states - correct_states) < 0.0001)
    if not correct.all(): 
        return False
 
    return True

if __name__ == '__main__':
    if run(verbose=True, doplots=True): 
        print '\n\nPassed test!!!'
    else: 
        print '\n\nFailed test...'
    