'''
Created on Dec. 3, 2013

Contains a test for the dynamic programming package

@author: Andre Dozier
'''

import dypro.dypro as dp 
import numpy as np 
# import matplotlib.pyplot as plt

testname = 'Miranda and Fackler (2002) Ex. 7.6.5'
def run(verbose=False, doplots=False): 
    # test from Miranda and Fackler (2002). Applied Computational Economics and Finance. p. 180
    
    # set up model parameters 
    state0 = np.array([1])
    alpha1  = 14.
    beta1   = 0.8
    alpha2  = 10.
    beta2   = 0.4
    maxcap  = 30
    delta   = 0.9
    n       = maxcap + 1 
    T       = 30
    tolerance = 0.1
    
    # random variable realizations 
    inflow  = np.array([0,1,2,3,4]).reshape((-1,1))
    inflowprob = [0.1,0.2,0.4,0.2,0.1]
    
    # set up the state and decision grids and time   
    stategrid = dp.LinGrid(0,maxcap,n)
    decisiongrid = dp.LinGrid(0,maxcap,n) 
    
    # maximize or minimize? 
    ismax = True

    # define the state equation
    def statefxn(s, x, t, inflowval):
        return s + inflowval - x
    
    # simulate the system, fixing decisions if need be
    def simfxn(s, x, t, inflowval): 
        snext = s + inflowval - x
        if snext < 0: 
            return (0, s + inflowval)
        elif snext > maxcap: 
            return (maxcap, s + inflowval - maxcap) 
        else: 
            return (snext, x)
    
    # define the value function 
    def valuefxn(s, snext, x, t, inflowval):
        if x > s: 
            return -np.inf
        else: 
            return alpha1 * (x ** beta1) + alpha2 * ((s - x) ** beta2)  
    
    # define the transition probabilities function 
    def transProb(s, nexts, x, t, rndvar_i): # p, x, t, w): 
        return inflowprob[rndvar_i]
            
    # solve the problem  
    opt_results = dp.solve_infinitetime(valuefxn, statefxn, stategrid, decisiongrid, 
                                  tolerance=tolerance, rndvar_space=inflow, 
                                  transitionP=transProb, ismax=ismax, 
                                  discount=delta, verbose=verbose)
    
    if verbose: 
        print 'opt_decisions_indices'
        print opt_results['opt_decisions_indices']
        
    # trace optimal paths
    t = list(np.linspace(0, T, T + 1))
 
    # simulate again to perform check
    trace_results = dp.trace_optimal_paths(stategrid, decisiongrid, t, state0,
                    opt_results, ismax=ismax, verbose=False)
                    
    # extract desired information             
    decisions = trace_results['trace_decisions']
    states = trace_results['trace_states']
    values = trace_results['trace_values']

    # if doplots: 
        # # stochastic timeseries generation
        # decisions, states, values, rndvars = \
         # dp.trace_optimal_paths(stategrid, decisiongrid, t, state0,
                    # opt_decisions_indices, opt_next_states_indices,
                    # opt_values,
                    # ismax=ismax, verbose=False,
                    # rndvar_space=inflow, transitionP=transProb, valuefxn=valuefxn, 
                    # simfxn=simfxn, nscenarios=10000)            
        # plt.close('all')

        # # fix variables to be 2D arrays (each column is a different stochastically generated timeseries)  
        # states = np.array([s.reshape((-1,)) for s in states])
        # decisions = np.array([d.reshape((-1,)) for d in decisions])
        # values = np.array([v.reshape((-1,)) for v in values])
        # meanstates = states.mean(axis=1) 
        # meandecisions = decisions.mean(axis=1)
        # meanvalues = values.mean(axis=1) 
        
        # if verbose: 
            # print 'mean decisions:'
            # print meandecisions
            # print 'mean states:'
            # print meanstates
            # print 'mean values'
            # print meanvalues 
            
        # # plot decisions against time 
        # plt.figure()
        # timearray = np.array(t[:-1]).reshape((-1,1))
        # plt.plot(timearray, decisions, 'k-', alpha=0.1)
        # plt.plot(timearray, meandecisions, 'k-', linewidth=3)
        # plt.xlabel('Time')
        # plt.ylabel('Optimal decision path')
        
        # # plot stock against time 
        # timearray = np.array(t).reshape((-1,1))
        # plt.figure()
        # plt.plot(timearray, states, 'k-', alpha=0.1)
        # plt.plot(timearray, meanstates, 'k-', linewidth=3)
        # plt.xlabel('Time')
        # plt.ylabel('Optimal state path')
        
        # # plot stock against time 
        # plt.figure()
        # plt.plot(timearray, values, 'k-', alpha=0.1)
        # plt.plot(timearray, meanvalues, 'k-', linewidth=3)
        # plt.xlabel('Time')
        # plt.ylabel('Optimal value path')

        # # plot iteration errors 
        # plt.figure()
        # plt.plot(maxerrors, 'k-',label='Errors')
        # plt.plot(np.arange(len(maxerrors)), np.ones(len(maxerrors)) * tolerance, 'k--',label='tolerance')
        # plt.xlabel('Iteration')
        # plt.ylabel('Error')

        # # show figures
        # plt.ion()
        # plt.show() 
                    
    # check optimal decisions
    correct_decisions = np.array([0,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2]).reshape((-1, 1))
    correct = (np.abs(decisions - correct_decisions) < 0.0001)
    if not correct.all(): 
        return False
          
    # check optimal values
    correct_states = np.array([1,3,4,5,6,7,8,9,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10]).reshape((-1, 1))
    correct = (np.abs(states - correct_states) < 0.0001)
    if not correct.all(): 
        return False
 
    return True

if __name__ == '__main__':
    if run(verbose=True, doplots=True): 
        print '\nPassed test!!!'
    else: 
        print '\nFailed test...'
    