'''
@author:    Andre Dozier
@copyright: 2014 Andre Dozier  
@summary:   Optimizes the C-BT system with the goal of estimating the added 
            value of storage or trade in the reservoir system
            
            This file was created in partial requirements for AREC 540 (Natural 
            Resources Economics) at Colorado State University in Fall 2013. 
'''
import os, sys
import numpy as np
from scipy import optimize
from numpy.random import random_sample 
import time, string, csv, argparse
from definitions import ScenarioType

def getparser():
    '''
    Retrieves the argument parser for this program 
    '''

    # default valuse 
    scendict = ScenarioType.__dict__
    scentypes = string.join([k + ' = %d' % scendict[k] for k in scendict if isinstance(scendict[k], int)],', ')
    scendefault = ScenarioType.get_name(ScenarioType.FreeMarket)

    # parser object
    parser = argparse.ArgumentParser(description='Optimize Single Dam Operations Using Dynamic Programming for Various Institutional Scenarios.')
    
    # files 
    parser.add_argument('outdir', metavar='output_directory',  
            help='Output directory for saved figures and optimal solutions.')
    parser.add_argument('inflowfile', metavar='inflowfile', type=str,  
            help='Tab-delimited input file containing three columns: timestep, midpoint of inflow bins for that timestep, and probabilities of that inflow realization (characterized by the bin).')
    
    # discretizations
    parser.add_argument('-m','--maxstor', default=470000., type=float,
            help='The maximum storage capacity. Default is 470,000 acre-feet (AF).')
    parser.add_argument('--init_state', default=250000., type=float,
            help='Initial storage at time 0.')
    parser.add_argument('-sp','--spill', default=30000., type=float,
            help='Spill capacity from the reservoir. Default is 30,000 acre-feet (AF) per period.')
    
    # water users 
    parser.add_argument('-s', '--scentype', default=scendefault, type=str, 
            help='Specifies which institution problem setup to solve. Possible values are: %s' % scentypes)
    parser.add_argument('-a','--alpha', type=float, nargs='*', default=None, 
            help='The fraction of water usage for each user. Must have the same number of elements as users (-u, --users flag) and must sum to 1.')
    
    # time horizon  
    parser.add_argument('--timestep', default='annual', type=str,
            help='The timestep at which this model runs: either "annual" or "seas"')
    parser.add_argument('-T', '--nperiods', default=50, type=int,
            help='The time horizon or the time period to simulate in the case of an infinite horizon DP model.')
    parser.add_argument('-i', '--infhorizon', action='store_const', default=False, const=True,
            help='If flag is passed to this program, the DP optimization problem will be solved using an infinite horizon.')
    
    # optimization options  
    parser.add_argument('-r', '--risk', default=0.05, type=float,
            help='The acceptable risk level to allow for "infeasible" solutions... In the reservoir problem, this means spill or emptying the reservoir.')
    parser.add_argument('-d', '--discount', default=1.0, type=float,
            help='The discount rate (equivalent to 1/(1+i) where i is the interest rate) used to discount future value. Should be between 0 and 1.')

    # scenario running  
    parser.add_argument('-n', '--nscen', default=100, type=int,
            help='The number of scenarios to simulate when tracing optimal paths with stochastic DP.')
    parser.add_argument('-j', '--justtrace', action='store_const', const=True, default=False, 
            help='If true, this will just trace the optimal results, assuming that the optimal result matrices have been saved to file.')
    
    # program options 
    parser.add_argument('-t', '--nthreads', default=1, type=int, 
            help='Specifies the number of threads to be utilized when running this dynamic program.')
    parser.add_argument('-o', '--stdoutfile', default=None, type=str, 
            help='Specifies an output file for printed standard output. If not specified, values are printed to the console.')
    parser.add_argument('-p', '--showplot', action='store_const', default=False, const=True,
            help='If flag is passed to this program, plots will be shown at the end of the optimization.')
    parser.add_argument('-v', '--verbose', action='store_const', default=False, const=True,
            help='If flag is passed to this program, debugging messages are printed to console.')
    
    return parser 


if __name__ == '__main__':
    # parse the arguments
    args = getparser().parse_args()

    # initial print output 
    print '\nStarting scenario in %s...' % args.outdir
    
    # check timestep
    if args.timestep == 'annual': 
        from annualdef import problemdef 
    elif args.timestep == 'seas': 
        from seasdef import problemdef 
    else: 
        raise Exception('Timestep %s not implemented. Please use "annual" or "seas".' % args.timestep)

    # get the scenario type
    def try_parse_int(s):
        try:
            return int(s)
        except ValueError:
            return s
    scendict = ScenarioType.__dict__
    if try_parse_int(args.scentype) in scendict.values(): 
        args.scentype = ScenarioType.get_name(try_parse_int(args.scentype))
    if not args.scentype in scendict: 
        raise Exception('Scenario type %s is not allowed. Possible values are: %s' % (args.scentype, scentypes)) 
    scentype = scendict[args.scentype]
    
    # output directory
    outdir = os.path.join(args.outdir) 
    if not os.path.exists(outdir):
        os.makedirs(outdir)    
    
    # get the scenario name
    scen_name = string.join(problemdef.abbrev, '_')
    scen_name += '_%s_' % (args.scentype)
    
    # switch standard output 
    if args.stdoutfile != None:
        outfile = os.path.join(outdir, scen_name + args.stdoutfile) 
        print 'Saving console output to %s...' % outfile
        stdout = sys.stdout
        stderr = sys.stderr
        sys.stderr = sys.stdout = open(outfile, 'w')
    
    # get demand constraints and allocations
    if args.discount < 0 or args.discount > 1: 
        raise Exception('The discount rate must be between 0 and 1.')
    
    # let user know which parameters are being run
    print '\n****************************************'
    print 'Running Scenario:  ', scen_name
    print 'Output Directory:  ', outdir
    print 'Institution Type:  ', args.scentype
    print 'List of Users:     ', problemdef.users
    print 'Number of Threads: ', args.nthreads
    print ''
    print 'Scenario Setup: '
    print '    Inflow file              =', args.inflowfile
    print '    Infinite Time Horizon    =', args.infhorizon
    print '    Number of Periods        =', args.nperiods
    print '    Number of Scenarios      =', args.nscen
    print '    Acceptable Risk          =', args.risk
    print '    Alpha                    =', args.alpha
    print '    Initial Storage (AF)     =', args.init_state
    print '    Maximum Storage (AF)     =', args.maxstor
    print '    Spill Capacity (AF)      =', args.spill
    print '****************************************\n'
    

    ########################### Solve the System   ######################### 
    
    problemdef.initialize(outdir, 
                     # scenario 
                     scen_name, scentype,  
                     # inflow 
                     args.inflowfile,
                     # spill
                     args.spill, 
                     # storage 
                     args.maxstor,
                     # optional variables 
                     # spill 
                     spill_step=None, 
                     # bank  
                     bank_percentage=0.2, withdrawal_loss=0.1,
                     # alloted percentage 
                     alpha=args.alpha)
    
    # get new optimal results, or load old ones 
    if args.justtrace: 
        problemdef.load()
    else: 
        problemdef.solve(# time horizon
            infhorizon=args.infhorizon, T=args.nperiods,
            # optimization options
            acceptablerisk=args.risk, discount=args.discount, 
            # state variables
            init_state=250000, 
            # program options  
            verbose=args.verbose, nthreads=args.nthreads)
        
    # trace optimal results
    problemdef.trace(args.nperiods, nscenarios=args.nscen, 
                     init_state=args.init_state, 
                     verbose=args.verbose, discount=args.discount)

    # plot 
    problemdef.plot(verbose=args.verbose, showplots=args.showplot)

    # finished 
    if args.stdoutfile != None:
        sys.stdout.close()
        sys.stdout = stdout
        sys.stderr = stderr
        print 'Finished scenario %s!!!\n' % os.path.dirname(outfile)

    else: 
        print 'Finished scenario %s!!!\n' % outdir















