'''
Author:    Andre Dozier
Date:      June 30, 2014
Purpose:   Run scenarios for the CBT system optimization in parallel  
'''
import os
import string 
import argparse 
import subprocess
import multiprocessing
from definitions import ScenarioType
    
# global variables 
rescap = 470000 
figdir = 'figures' 
datadir = 'data' 
inflowprefix = 'inflowprob'

# global functions 
def solve(scen):
    return subprocess.call(['python','scenario.py'] + scen.args())

class Scenario:
    '''
    A class that builds the argument list for running lots of CBT system 
    optimization scenarios  
    '''

    def __init__(self, scentype, timestep, 
                 maxstor_shift=0.0, spill=None,  
                 alpha=None,
                 risk=0.0, discount=0.9524, banking=False, 
                 mean_shift=0, std_shift=0, nscen=100, justtrace=False):
        
        self.timestep = timestep 
        
        self.maxstor_shift = maxstor_shift 
        self.spill = spill if spill != None else 30000

        self.scentype = scentype
        self.alpha = alpha
        self.userabbrev = ['Ag', 'Muni'] # matches order in alpha

        self.risk = risk 
        self.discount = discount 
        self.banking = banking 
         
        self.mean_shift = mean_shift 
        self.std_shift = std_shift
        self.nscen = nscen 

        self.justtrace = justtrace
    
    def directory(self):
        # first subdirectory
        scenname = '' 
        if self.maxstor_shift != 0.0: 
            scenname += 'Stor%+05.1f' % (self.maxstor_shift * 100.)
        if self.mean_shift != 0: 
            scenname += 'Mean%2d' % int(self.mean_shift)
        if self.std_shift != 0: 
            scenname += 'Std%2d' % int(self.std_shift) 
        if scenname == '': 
            scenname = 'Original'
        
        # second subdirectory
        if self.alpha == None: 
            dirname = 'FreeMarket'
        else: 
            dirname = string.join([self.userabbrev[i] + str(int(self.alpha[i] * 100)) for i in range(len(self.alpha))], '')
        if self.banking: 
            dirname += '_Bank'
            
        return os.path.join(figdir + '_' + self.timestep, scenname, dirname)  

    def inflowfile(self):
        mean_shift = '_Mean%2d' % int(self.mean_shift) if self.mean_shift != 0 else ''
        var_shift = '_Std%2d' % int(self.std_shift) if self.std_shift != 0 else ''
        return os.path.join(datadir, self.timestep, inflowprefix + mean_shift + var_shift + '.txt') 
    
    def args(self):
        # output directory 
        outdir = self.directory()
        
        # output console file
        consfile = 'console_output.txt'
        
        # max storage
        maxstor = rescap + self.maxstor_shift * rescap
        
        # list of arguments
        return [outdir, self.inflowfile(),  
                '--timestep', self.timestep,
                '--maxstor', str(maxstor), 
                '--init_state', str(0.0 if abs(maxstor - 0.0) < 249999.99 else 250000.0),
                '--spill', str(self.spill),
                '--scentype', str(self.scentype)] + ([
                '--alpha'] + [str(a) for a in self.alpha] if self.alpha != None else []) + [
                '--risk', str(self.risk),
                '--discount', str(self.discount),
                '--nscen', str(self.nscen),
                '--stdoutfile', consfile] + ([
                '--justtrace'] if self.justtrace else [])

def build_args(tstep, maxstor_shifts, mean_shifts, std_shifts, justtrace):
    allargs = [] 
    for maxstor_shift in maxstor_shifts: 
        for mean_shift in mean_shifts: 
            for std_shift in std_shifts:
                risk = 1.0 if maxstor_shift == -1 else 0.0  
                allargs += [Scenario(ScenarioType.FreeMarket, tstep, alpha=None, maxstor_shift=maxstor_shift, mean_shift=mean_shift, std_shift=std_shift, risk=risk, justtrace=justtrace),
                            Scenario(ScenarioType.PriorApprop, tstep, alpha=[0.85,0.15], maxstor_shift=maxstor_shift, mean_shift=mean_shift, std_shift=std_shift, risk=risk, justtrace=justtrace),
#                            Scenario(ScenarioType.PriorApprop, tstep, alpha=[0.2,0.8], maxstor_shift=maxstor_shift, mean_shift=mean_shift, std_shift=std_shift, risk=risk, justtrace=justtrace),
                            Scenario(ScenarioType.PriorApprop, tstep, alpha=[0.64,0.36], maxstor_shift=maxstor_shift, mean_shift=mean_shift, std_shift=std_shift, risk=risk, justtrace=justtrace)]
    return allargs
    
if __name__ == '__main__':
    # special imports
    
    # parse arguments
    parser = argparse.ArgumentParser(description='Run scenarios for the CBT system optimization in parallel')
    parser.add_argument('--timestep', type=str, default='annual', 
         help='The timestep to use. Possible values are "annual" or "seas". Default is "seas"')
    parser.add_argument('-np', '--poolsize', type=int, default=max(multiprocessing.cpu_count() - 1, 1), 
         help='The number of processors used to run all different scenarios, default is Max(1, Number of CPUs minus 1).')
    parser.add_argument('--justtest', action='store_const', default=False, const=True, 
         help='A flag specifying just to test the package (i.e., run the baseline scenario), not to run all the scenarios.')
    parser.add_argument('--justprintargs', action='store_const', default=False, const=True, 
         help='A flag specifying just to print the arguments for all the programs.')
    parser.add_argument('--justtrace', action='store_const', default=False, const=True, 
         help='A flag specifying just to trace the optimal scenarios.')
    args = parser.parse_args()
    
    # checks 
    assert args.timestep.lower() == 'annual' or args.timestep.lower() == 'seas', 'Possible values for the timestep input parameter is "annual" or "seas".'
    
    # arguments for all desired plots
    if args.justtest: 
        allargs = build_args(args.timestep, [0], [0], [0], args.justtrace) # for testing
    else: 
        allargs = build_args(args.timestep, [0, 0.125, 0.25, -1], [0, 10, 20], [0], args.justtrace)
    
    # run scenarios on 6 different processors (if 8 processors exist)
    if args.justprintargs: 
        for s in allargs:
            print string.join(s.args(), ' ') 
    else:              
        pool = multiprocessing.Pool(processes=args.poolsize)
        r = ['Failed' if r != 0 else 'Success' for r in pool.map(solve, allargs)]
        print 'Finished processes : ', r
    
    
    
    
    
    